# Generated by Django 5.2 on 2025-07-04 17:07

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0108_remove_record_sheet_remove_sheetproject_sheet_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="record",
            name="input_of_runs",
            field=models.ManyToManyField(
                related_name="input_records", to="lamindb.run"
            ),
        ),
        migrations.AlterField(
            model_name="record",
            name="run",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="output_records",
                to="lamindb.run",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="recordrecord",
            unique_together={("record", "feature", "value")},
        ),
        migrations.AlterUniqueTogether(
            name="recordrun",
            unique_together={("record", "feature", "value")},
        ),
        migrations.AlterUniqueTogether(
            name="recordulabel",
            unique_together={("record", "feature", "value")},
        ),
        migrations.AddField(
            model_name="record",
            name="values_artifacts",
            field=models.ManyToManyField(
                related_name="linked_in_records",
                through="lamindb.RecordArtifact",
                to="lamindb.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="recordartifact",
            name="value",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="links_linked_in_record",
                to="lamindb.artifact",
            ),
        ),
        migrations.CreateModel(
            name="ArtifactRecord",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "artifact",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links_record",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifactrecord",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "record",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="links_artifact",
                        to="lamindb.record",
                    ),
                ),
            ],
            options={
                "unique_together": {("artifact", "record", "feature")},
            },
            bases=(models.Model, lamindb.models.sqlrecord.IsLink),
        ),
        migrations.AlterField(
            model_name="record",
            name="artifacts",
            field=models.ManyToManyField(
                related_name="records",
                through="lamindb.ArtifactRecord",
                to="lamindb.artifact",
            ),
        ),
        migrations.AddField(
            model_name="artifactrecord",
            name="feature_ref_is_name",
            field=lamindb.base.fields.BooleanField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactrecord",
            name="label_ref_is_name",
            field=lamindb.base.fields.BooleanField(blank=True, default=None, null=True),
        ),
    ]



import lamindb as ln
import pytest


def test_run():
    transform = ln.Transform(key="My transform")
    with pytest.raises(ValueError) as error:
        ln.Run(transform)
    assert (
        error.exconly()
        == "ValueError: Please save transform record before creating a run"
    )
    transform.save()
    run = ln.Run(transform).save()
    assert run.reference is None
    assert run.reference_type is None
    run2 = ln.Run(transform, reference="test1", reference_type="test2").save()
    assert run2.reference == "test1"
    assert run2.reference_type == "test2"
    assert run.uid != run2.uid

    report_artifact = ln.Artifact("README.md", description="report of run2").save()
    run2.report = report_artifact
    environment = ln.Artifact("CONTRIBUTING.md", description="env of run2").save()
    run2.environment = environment

    run2.delete(permanent=True)

    # test deletion of run including attached artifacts
    assert ln.Artifact.objects.filter(uid=report_artifact.uid).exists() is False
    assert ln.Artifact.objects.filter(uid=environment.uid).exists() is False

    transform.delete(permanent=True)

    assert ln.Run.filter(uid=run.uid).count() == 0


def test_edge_cases():
    with pytest.raises(ValueError) as error:
        ln.Run(1, 2)
    assert error.exconly() == "ValueError: Only one non-keyword arg allowed: transform"
    with pytest.raises(TypeError) as error:
        ln.Run()
    assert error.exconly() == "TypeError: Pass transform parameter"



"""Errors.

.. autosummary::
   :toctree: .

   ValidationError
   InvalidArgument
   DoesNotExist
   NotebookNotSaved
   UnknownStorageLocation
   MissingContextUID
   UpdateContext
   IntegrityError
   FieldValidationError
   SQLRecordNameChangeIntegrityError
   NoWriteAccess

"""

# inheriting from SystemExit has the sole purpose of suppressing
# the traceback - this isn't optimal but the current best solution
# https://laminlabs.slack.com/archives/C04A0RMA0SC/p1726856875597489


class ValidationError(Exception):
    """Validation error."""

    pass


class InvalidArgument(Exception):
    """Invalid method or function argument."""

    pass


class TrackNotCalled(Exception):
    """`ln.track()` wasn't called."""

    pass


class NotebookNotSaved(Exception):
    """Notebook wasn't saved."""

    pass


class UnknownStorageLocation(Exception):
    """Path is not contained in any known storage location."""

    pass


# equivalent to Django's DoesNotExist
# and SQLAlchemy's NoResultFound
class DoesNotExist(Exception):
    """No record found."""

    pass


class InconsistentKey(Exception):
    """Inconsistent transform or artifact `key`."""

    pass


class SQLRecordNameChangeIntegrityError(Exception):
    """Custom exception for name change errors."""

    pass


class FieldValidationError(Exception):
    """Field validation error."""

    pass


# -------------------------------------------------------------------------------------
# run context
# -------------------------------------------------------------------------------------


class IntegrityError(Exception):
    """Integrity error.

    For instance, it's not allowed to delete artifacts outside managed storage
    locations.
    """

    pass


class MissingContextUID(SystemExit):
    """User didn't define transform settings."""

    pass


class UpdateContext(SystemExit):
    """Transform settings require update."""

    pass


# -------------------------------------------------------------------------------------
# record
# -------------------------------------------------------------------------------------


class NoWriteAccess(Exception):
    """No write access to a space."""

    pass



"""Examples for MLCommons Croissant files, which are used to store metadata about datasets.

.. autosummary::
   :toctree: .

   mini_immuno

"""

import json
from pathlib import Path


def mini_immuno(n_files: int = 1) -> list[Path]:
    """Return paths to the mini immuno dataset and its metadata as a Croissant file.

    Args:
        n_files: Number of files inside the croissant file. Default is 1.

    Example

        ::

            croissant_path, dataset1_path = ln.examples.croissant.mini_immuno()
    """
    from ..datasets import file_mini_csv
    from ..datasets.mini_immuno import get_dataset1

    adata = get_dataset1(otype="AnnData")
    dataset1_path = Path("mini_immuno.anndata.zarr")
    adata.write_zarr(dataset1_path)
    orig_croissant_path = (
        Path(__file__).parent / "mini_immuno.anndata.zarr_metadata.json"
    )
    with open(orig_croissant_path, encoding="utf-8") as f:
        data = json.load(f)
    if n_files == 2:
        dataset2_path = file_mini_csv()
        data["distribution"].append(
            {
                "@type": "sc:FileObject",
                "@id": "mini.csv",
                "name": "mini.csv",
                "encodingFormat": "text/csv",
            }
        )
    croissant_path = Path("mini_immuno.anndata.zarr_metadata.json")
    with open(croissant_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2)
    result: list[Path] = [croissant_path, dataset1_path]
    if n_files == 1:
        return result
    result.append(dataset2_path)
    return result



# ruff: noqa: F811

import datetime
from pathlib import Path

import bionty as bt
import lamindb as ln
import pytest
from _dataset_fixtures import (  # noqa
    get_mini_csv,
)
from lamindb.errors import DoesNotExist, ValidationError
from lamindb.examples.datasets import small_dataset1
from lamindb.models._feature_manager import describe_features
from lamindb.models._label_manager import format_rich_tree
from lamindb.models.artifact import add_labels


@pytest.fixture(scope="module")
def adata():
    adata = ln.core.datasets.anndata_with_obs()
    # add another column
    adata.obs["cell_type_by_expert"] = adata.obs["cell_type"]
    adata.obs.loc["obs0", "cell_type_by_expert"] = "B cell"
    return adata


def test_features_add():
    df = small_dataset1(otype="DataFrame")
    artifact = ln.Artifact.from_dataframe(df, description="test dataset").save()
    with pytest.raises(ValidationError) as err:
        artifact.features.add_values({"perturbation": df.perturbation.unique()})
    assert (
        err.exconly()
        == """lamindb.errors.ValidationError: These keys could not be validated: ['perturbation']
Here is how to create a feature:

  ln.Feature(name='perturbation', dtype='cat').save()"""
    )

    ln.Feature(name="perturbation", dtype="cat").save()
    ln.ULabel.from_values(["DMSO", "IFNG"], create=True).save()
    artifact.features.add_values({"perturbation": df.perturbation.unique()})
    assert artifact in ln.Artifact.filter(perturbation__isnull=False)
    assert artifact not in ln.Artifact.filter(perturbation__isnull=True)

    artifact.delete(permanent=True)
    ln.ULabel.filter().delete()
    ln.Feature.filter().delete()


def test_features_add_external():
    df = small_dataset1(otype="DataFrame")
    artifact = ln.Artifact.from_dataframe(df, description="test dataset").save()

    species = ln.Feature(name="species", dtype="str").save()
    split = ln.Feature(name="split", dtype="str").save()
    schema = ln.Schema([species, split]).save()

    with pytest.raises(ValidationError) as e:
        artifact.features.add_values({"doesnot": "exist"}, schema=schema)
    assert "column 'split' not in dataframe" in str(e.value)

    artifact.features.add_values({"species": "bird", "split": "train"}, schema=schema)
    artifact.save()

    assert artifact.features.get_values() == {"species": "bird", "split": "train"}

    artifact.delete(permanent=True)
    schema.delete(permanent=True)
    ln.Feature.filter().delete()


# below the test for annotating with feature values
def test_features_add_remove(adata):
    artifact = ln.Artifact.from_anndata(adata, description="test").save()
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values({"experiment": "Experiment 1"})
    assert error.exconly().startswith(
        "lamindb.errors.ValidationError: These keys could not be validated:"
    )
    ln.Feature(name="experiment", dtype="cat").save()
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values({"experiment": "Experiment 1"})
    assert error.exconly().startswith(
        "lamindb.errors.ValidationError: These values could not be validated:"
    )
    experiment_label = ln.ULabel(name="Experiment 1").save()
    # add the label without the feature first
    artifact.ulabels.add(experiment_label)
    assert artifact.links_ulabel.get().ulabel.name == "Experiment 1"
    assert artifact.links_ulabel.get().feature is None

    # now add the label with the feature and make sure that it has the feature annotation
    artifact.features.add_values({"experiment": "Experiment 1"})
    assert artifact.links_ulabel.get().ulabel.name == "Experiment 1"
    assert artifact.links_ulabel.get().feature.name == "experiment"
    # repeat
    artifact.features.add_values({"experiment": "Experiment 1"})
    assert artifact.links_ulabel.get().ulabel.name == "Experiment 1"

    # numerical feature
    temperature = ln.Feature(name="temperature", dtype="cat").save()
    with pytest.raises(TypeError) as error:
        artifact.features.add_values({"temperature": 27.2})
    assert error.exconly().startswith(
        "TypeError: Value for feature 'temperature' with dtype 'cat' must be a string or record"
    )
    temperature.dtype = "num"
    temperature.save()
    artifact.features.add_values({"temperature": 27.2})
    assert artifact._feature_values.first().value == 27.2

    # datetime feature
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values({"date_of_experiment": "2024-12-01"})
    assert (
        error.exconly()
        == """lamindb.errors.ValidationError: These keys could not be validated: ['date_of_experiment']
Here is how to create a feature:

  ln.Feature(name='date_of_experiment', dtype='date').save()"""
    )

    ln.Feature(name="date_of_experiment", dtype="date").save()
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values({"date_of_experiment": "Typo2024-12-01"})
    assert (
        error.exconly()
        == """lamindb.errors.ValidationError: Expected dtype for 'date_of_experiment' is 'date', got 'cat ? str'"""
    )
    artifact.features.add_values({"date_of_experiment": "2024-12-01"})

    ln.Feature(name="datetime_of_experiment", dtype="datetime").save()
    artifact.features.add_values({"datetime_of_experiment": "2024-12-01 00:00:00"})

    # bionty feature
    mouse = bt.Organism.from_source(name="mouse")
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values({"organism": mouse})
    assert (
        error.exconly()
        == """lamindb.errors.ValidationError: These keys could not be validated: ['organism']
Here is how to create a feature:

  ln.Feature(name='organism', dtype='cat[bionty.Organism]').save()"""
    )
    ln.Feature(name="organism", dtype="cat[bionty.Organism]").save()
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values({"organism": mouse})
    assert (
        # ensure the label is saved
        error.exconly().startswith("lamindb.errors.ValidationError: Please save")
    )
    mouse.save()
    artifact.features.add_values({"organism": mouse})
    assert artifact.organisms.get().name == "mouse"

    # lists of records
    diseases = bt.Disease.from_values(
        ["MONDO:0004975", "MONDO:0004980"], field=bt.Disease.ontology_id
    )
    ln.save(diseases)
    ln.Feature(name="disease", dtype="cat[bionty.Disease]").save()
    artifact.features.add_values({"disease": diseases})
    assert len(artifact.diseases.filter().all()) == 2

    # big dictionary of everything
    features = {
        "experiment": [  # we're testing iterable annotation here
            "Experiment 2",
            "Experiment 1",
        ],
        "project": "project_1",
        "is_validated": True,
        "cell_type_by_expert": "T cell",
        "temperature": 100.0,
        "donor": "U0123",
    }
    with pytest.raises(ValidationError) as error:
        artifact.features.add_values(features)
    assert (
        error.exconly()
        == """\
lamindb.errors.ValidationError: These keys could not be validated: ['project', 'is_validated', 'cell_type_by_expert', 'donor']
Here is how to create a feature:

  ln.Feature(name='project', dtype='cat ? str').save()
  ln.Feature(name='is_validated', dtype='bool').save()
  ln.Feature(name='cell_type_by_expert', dtype='cat ? str').save()
  ln.Feature(name='donor', dtype='cat ? str').save()"""
    )

    ln.Feature(name="project", dtype=ln.ULabel).save()
    ln.Feature(name="is_validated", dtype=bool).save()
    ln.Feature(name="cell_type_by_expert", dtype=bt.CellType).save()
    ln.Feature(name="donor", dtype=ln.ULabel).save()

    with pytest.raises(ValidationError) as error:
        artifact.features.add_values(features)
        error_msg = error.exconly()

        assert (
            "lamindb.errors.ValidationError: These values could not be validated:"
            in error_msg
        )
        assert "Here is how to create records for them:" in error_msg

        expected_values = {
            "ULabel": ["project_1", "U0123", "Experiment 2"],
            "bionty.CellType": ["T cell"],
        }

        for key, values in expected_values.items():
            assert f"'{key}':" in error_msg
            for value in values:
                assert value in error_msg
            assert f"{key.split('.')[-1]}.from_values(" in error_msg

        assert "create=True).save()" in error_msg

    ln.ULabel.from_values(["Experiment 2", "project_1", "U0123"], create=True).save()
    bt.CellType.from_source(name="T cell").save()
    print("validate", bt.CellType.validate(["T cell"]))

    artifact.features.add_values(features)
    assert set(artifact._feature_values.all().values_list("value", flat=True)) == {
        27.2,
        True,
        100.0,
        "2024-12-01",
        "2024-12-01T00:00:00",
    }

    assert ln.Artifact.get(_feature_values__value=27.2)

    assert artifact.features.get_values() == {
        "disease": {"Alzheimer disease", "atopic eczema"},
        "experiment": {"Experiment 1", "Experiment 2"},
        "project": "project_1",
        "cell_type_by_expert": "T cell",
        "donor": "U0123",
        "organism": "mouse",
        "is_validated": True,
        "temperature": {27.2, 100.0},
        "date_of_experiment": datetime.date(2024, 12, 1),
        "datetime_of_experiment": datetime.datetime(2024, 12, 1, 0, 0, 0),
    }
    # hard to test because of italic formatting
    tree = describe_features(artifact)
    format_rich_tree(tree)
    assert tree.children[0].label.plain == "Linked features"
    assert len(tree.children[0].children[0].label.columns) == 3
    assert len(tree.children[0].children[0].label.rows) == 10
    assert tree.children[0].children[0].label.columns[0]._cells == [
        "cell_type_by_expert",
        "disease",
        "donor",
        "experiment",
        "organism",
        "project",
        "date_of_experiment",
        "datetime_of_experiment",
        "is_validated",
        "temperature",
    ]
    types = [i.plain for i in tree.children[0].children[0].label.columns[1]._cells]
    assert types == [
        "cat[bionty.CellType]",
        "cat[bionty.Disease]",
        "cat[ULabel]",
        "cat[ULabel]",
        "cat[bionty.Organism]",
        "cat[ULabel]",
        "date",
        "datetime",
        "bool",
        "num",
    ]
    assert tree.children[0].children[0].label.columns[2]._cells == [
        "T cell",
        "Alzheimer disease, atopic eczema",
        "U0123",
        "Experiment 1, Experiment 2",
        "mouse",
        "project_1",
        "2024-12-01",
        "2024-12-01 00:00:00",
        "True",
        "27.2, 100.0",
    ]
    # repeat
    artifact.features.add_values(features)
    assert set(artifact._feature_values.all().values_list("value", flat=True)) == {
        27.2,
        True,
        100.0,
        "2024-12-01",
        "2024-12-01T00:00:00",
    }

    with pytest.raises(ln.errors.InvalidArgument) as error:
        ln.Artifact.filter(temperature_with_typo=100.0, project="project_1").one()
    assert error.exconly().startswith(
        "lamindb.errors.InvalidArgument: You can query either by available fields:"
    )

    ln.Artifact.filter(temperature=100.0)
    ln.Artifact.filter(project="project_1")
    ln.Artifact.filter(is_validated=True)
    ln.Artifact.filter(temperature=100.0, project="project_1", donor="U0123").one()
    # for bionty
    assert (
        artifact == ln.Artifact.filter(disease=diseases[0]).one()
    )  # value is a record
    assert (
        artifact == ln.Artifact.filter(disease="Alzheimer disease").one()
    )  # value is a string
    assert artifact == ln.Artifact.filter(disease__contains="Alzheimer").one()

    # test not finding the ULabel
    with pytest.raises(DoesNotExist) as error:
        ln.Artifact.filter(project="project__1")
    assert error.exconly().startswith(
        "lamindb.errors.DoesNotExist: Did not find a ULabel matching"
    )

    # test comparator
    assert artifact == ln.Artifact.filter(experiment__contains="ment 1").one()
    # due to the __in comparator, we get the same artifact twice below
    # print(ln.Artifact.to_dataframe(features=["experiment"]))
    # print(ln.Artifact.filter(experiment__contains="Experi").to_dataframe(features=["experiment"]))
    assert len(ln.Artifact.filter(experiment__contains="Experi").all()) == 2
    assert ln.Artifact.filter(temperature__lt=21).one_or_none() is None
    assert len(ln.Artifact.filter(temperature__gt=21).all()) >= 1

    # test remove_values
    artifact.features.remove_values("date_of_experiment")
    alzheimer = bt.Disease.get(name="Alzheimer disease")
    artifact.features.remove_values("disease", value=alzheimer)
    values = artifact.features.get_values()
    assert "date_of_experiment" not in values
    assert "Alzheimer disease" not in values["disease"]

    # test annotate with dictionaries multiple times
    ln.Feature(name="study_metadata", dtype=dict).save()
    artifact.features.add_values({"study_metadata": {"detail1": "123", "detail2": 1}})

    # delete everything we created
    artifact.delete(permanent=True)
    ln.ULabel.filter().delete()
    ln.Schema.filter().delete()
    ln.Feature.filter().delete()
    bt.Gene.filter().delete()
    bt.Organism.filter().delete()
    bt.Disease.filter().delete()


# most underlying logic here is comprehensively tested in test_context
def test_params_add():
    path = Path("mymodel.pt")
    path.touch()
    artifact = ln.Artifact("mymodel.pt", kind="model", description="hello").save()
    ln.Feature(name="learning_rate", dtype="float").save()
    ln.Feature(name="quantification", dtype="dict").save()
    artifact.features.add_values({"learning_rate": 0.01})
    artifact.features.add_values(
        {
            "quantification": {
                "name": "mcquant",
                "container": "labsyspharm/quantification",
            }
        }
    )
    assert artifact.features.get_values() == {
        "learning_rate": 0.01,
        "quantification": {
            "name": "mcquant",
            "container": "labsyspharm/quantification",
        },
    }
    # test describe params
    tree = describe_features(artifact)
    assert tree.label.plain == "Artifact .pt"
    assert tree.children[0].label.plain == "Linked features"
    assert len(tree.children[0].children[0].label.columns) == 3
    assert tree.children[0].children[0].label.columns[0]._cells == [
        "learning_rate",
        "quantification",
    ]
    assert tree.children[0].children[0].label.columns[1]._cells[0].plain == "float"
    assert tree.children[0].children[0].label.columns[1]._cells[1].plain == "dict"
    assert tree.children[0].children[0].label.columns[2]._cells == [
        "0.01",
        "{'name': 'mcquant', 'container': 'labsyspharm/quantification'}",
    ]
    artifact.describe()
    artifact.delete(permanent=True)
    path.unlink()


def test_labels_add(adata):
    label = ln.ULabel(name="Experiment 1")
    artifact = ln.Artifact.from_anndata(adata, description="test")
    artifact.save()
    experiment = ln.Feature(name="experiment", dtype="cat")
    with pytest.raises(ValueError) as error:
        artifact.labels.add("experiment_1", experiment)
    assert (
        error.exconly()
        == "ValueError: Please pass a record (a `SQLRecord` object), not a string, e.g.,"
        " via: label = ln.ULabel(name='experiment_1')"
    )
    with pytest.raises(ValidationError) as error:
        artifact.labels.add(label, experiment)
    assert "not validated. If it looks correct: record.save()" in error.exconly()
    label.save()
    with pytest.raises(TypeError) as error:
        artifact.labels.add(label, "experiment 1")
    with pytest.raises(ValidationError) as error:
        artifact.labels.add(label, feature=experiment)
    assert (
        error.exconly()
        == "lamindb.errors.ValidationError: Feature not validated. If it looks"
        " correct: ln.Feature(name='experiment', type='cat[ULabel]').save()"
    )
    experiment.save()

    # try to pass list of length zero
    artifact.labels.add([], feature=experiment)
    # now pass a single label
    artifact.labels.add(label, feature=experiment)
    # check that the feature was updated with type = "ULabel"
    feature = ln.Feature.get(name="experiment")
    assert feature.dtype == "cat[ULabel]"
    with pytest.raises(TypeError):
        experiments = artifact.labels.get("experiment")
    # check that the label is there, it's exactly one label with name "Experiment 1"
    experiments = artifact.labels.get(experiment)
    assert experiments.one().name == "Experiment 1"

    # try adding the same label again, nothing should happen
    artifact.labels.add(label, feature=experiment)
    # check that the label is there, it's exactly one label with name "Experiment 1"
    experiments = artifact.labels.get(experiment)
    assert experiments.get().name == "Experiment 1"

    # running from_values to load validated label records under the hood
    experiment = ln.Feature(name="experiment_with_reg", dtype="cat[ULabel]").save()
    ln.ULabel(name="Experiment 2").save()
    artifact.labels.add("Experiment 2", experiment)
    experiments = artifact.labels.get(experiment)
    assert experiments.get().name == "Experiment 2"

    # now, try adding a new label
    project = ln.ULabel(name="project 1").save()
    ln.Feature(name="project", dtype="cat").save()
    features = ln.Feature.lookup()
    artifact.labels.add(project, feature=features.project)
    # check that the label is there, it's exactly one label with name "Experiment 1"
    projects = artifact.labels.get(features.project)
    assert projects.get().name == "project 1"

    # test add_from
    adata2 = adata.copy()
    adata2.uns["mutated"] = True
    artifact2 = ln.Artifact(adata2, description="My new artifact").save()

    artifact2.labels.add_from(artifact)
    experiments = artifact2.labels.get(experiment)
    assert experiments.get().name == "Experiment 2"

    artifact2.delete(permanent=True)
    artifact.delete(permanent=True)
    ln.Schema.filter().delete()
    ln.Feature.filter().delete()
    ln.ULabel.filter().delete()


def test_add_labels_using_anndata(adata):
    organism = bt.Organism.from_source(name="mouse")
    cell_types = [bt.CellType(name=name) for name in adata.obs["cell_type"].unique()]
    ln.save(cell_types)
    inspector = bt.CellType.inspect(adata.obs["cell_type_by_expert"].unique())
    ln.save([bt.CellType(name=name) for name in inspector.non_validated])
    cell_types_from_expert = bt.CellType.from_values(
        adata.obs["cell_type_by_expert"].unique()
    )
    actual_tissues = [bt.Tissue(name=name) for name in adata.obs["tissue"].unique()]
    organoid = ln.ULabel(name="organoid")
    tissues = actual_tissues + [organoid]
    ln.save(tissues)

    bt.settings.organism = "human"

    # clean up DB state
    organism_feature = ln.Feature.filter(name="organism").one_or_none()
    if organism_feature is not None:
        organism_feature.delete()
    artifact = ln.Artifact.filter(description="Mini adata").one_or_none()
    if artifact is not None:
        artifact.delete(permanent=True, storage=True)
    ln.Schema.filter().delete()

    # try to construct without registering metadata features
    artifact = ln.Artifact.from_anndata(adata, description="Mini adata")
    if not artifact._state.adding:
        artifact.delete(permanent=True)  # make sure we get a fresh one
        artifact = ln.Artifact.from_anndata(adata, description="Mini adata")
    # add feature set without saving file
    feature_name_feature = ln.Feature(name="feature name", dtype="cat[ULabel]").save()
    schema = ln.Schema(features=[feature_name_feature])
    with pytest.raises(ValueError) as error:
        artifact.features._add_schema(schema, slot="random")
    assert (
        error.exconly()
        == "ValueError: Please save the artifact or collection before adding a feature"
        " set!"
    )

    # now register features we want to validate
    # (we are not interested in cell_type_id, here)
    ln.save(
        ln.Feature.from_dataframe(
            adata.obs[["cell_type", "tissue", "cell_type_by_expert", "disease"]]
        )
    )
    artifact = ln.Artifact.from_anndata(adata, description="Mini adata")
    ln.Feature(name="organism", dtype="cat[bionty.Organism]").save()
    features = ln.Feature.lookup()
    with pytest.raises(ValueError) as error:
        artifact.labels.add(organism, feature=features.organism)
    assert (
        error.exconly()
        == "ValueError: Please save the artifact/collection before adding a label!"
    )
    artifact.save()

    # link features
    artifact.features._add_set_from_anndata(var_field=bt.Gene.ensembl_gene_id)

    # check the basic construction of the feature set based on obs
    schema_obs = artifact.feature_sets.filter(
        itype="Feature", _links_artifact__slot="obs"
    ).one()
    assert schema_obs.n == 4
    assert "organism" not in schema_obs.features.to_list("name")

    # now, we add organism and run checks
    features = ln.Feature.lookup()
    with pytest.raises(ln.errors.ValidationError):
        artifact.labels.add(organism, feature=features.organism)
    organism.save()
    artifact.labels.add(organism, feature=features.organism)
    organism_link = artifact.links_organism.first()
    assert organism_link.organism.name == "mouse"
    assert organism_link.feature.name == "organism"
    feature = ln.Feature.get(name="organism")
    assert feature.dtype == "cat[bionty.Organism]"
    schema_obs = artifact.feature_sets.filter(
        itype="Feature", _links_artifact__slot="obs"
    ).one()
    assert schema_obs.n == 4

    # now we add cell types & tissues and run checks
    ln.Feature(name="cell_type", dtype="cat").save()
    ln.Feature(name="cell_type_by_expert", dtype="cat").save()
    ln.Feature(name="tissue", dtype="cat").save()
    add_labels(artifact, cell_types, feature=features.cell_type, from_curator=True)
    add_labels(
        artifact,
        cell_types_from_expert,
        feature=features.cell_type_by_expert,
        from_curator=True,
    )
    with pytest.raises(ValidationError) as err:
        add_labels(artifact, tissues, feature=features.tissue, from_curator=True)
    assert (
        err.exconly()
        == "lamindb.errors.ValidationError: Label type ULabel is not valid for Feature(name='tissue', dtype='cat[bionty.Tissue]'), consider updating to dtype='cat[bionty.Tissue|ULabel]'"
    )
    tissue = ln.Feature.get(name="tissue")
    tissue.dtype = "cat[bionty.Tissue|ULabel]"
    tissue.save()
    add_labels(artifact, tissues, feature=tissue, from_curator=True)
    feature = ln.Feature.get(name="cell_type")
    assert feature.dtype == "cat[bionty.CellType]"
    feature = ln.Feature.get(name="cell_type_by_expert")
    assert feature.dtype == "cat[bionty.CellType]"
    feature = ln.Feature.get(name="tissue")
    assert feature.dtype == "cat[bionty.Tissue|ULabel]"
    diseases = [ln.ULabel(name=name) for name in adata.obs["disease"].unique()]
    ln.save(diseases)
    add_labels(artifact, diseases, feature=features.disease, from_curator=True)
    df = artifact.features.slots["obs"].features.to_dataframe()
    assert set(df["name"]) == {
        "cell_type",
        "disease",
        "tissue",
        "cell_type_by_expert",
    }
    assert set(df["dtype"]) == {
        "cat[bionty.CellType]",
        "cat[ULabel]",
        "cat[bionty.Tissue|ULabel]",
    }

    # now, let's add another feature to ext
    experiment_1 = ln.ULabel(name="experiment_1").save()
    ln.Feature(name="experiment", dtype="cat").save()
    features = ln.Feature.lookup()
    artifact.labels.add(experiment_1, feature=features.experiment)
    # TODO: replace the following with an updated test
    # df = artifact.features["external"].to_dataframe()
    # assert set(df["name"]) == {
    #     "organism",
    #     "experiment",
    # }
    # assert set(df["dtype"]) == {"cat[bionty.Organism]", "cat[ULabel]"}

    assert set(artifact.labels.get(features.experiment).to_list("name")) == {
        "experiment_1"
    }
    assert set(artifact.labels.get(features.disease).to_list("name")) == {
        "chronic kidney disease",
        "Alzheimer disease",
        "liver lymphoma",
        "cardiac ventricle disorder",
    }
    assert set(artifact.labels.get(features.organism).to_list("name")) == {"mouse"}
    assert set(
        artifact.labels.get(features.tissue)["bionty.Tissue"].to_list("name")
    ) == {
        "liver",
        "heart",
        "kidney",
        "brain",
    }
    assert set(artifact.labels.get(features.tissue)["ULabel"].to_list("name")) == {
        "organoid",
    }
    # currently, we can't stratify the two cases below
    assert set(artifact.labels.get(features.cell_type).to_list("name")) == {
        "T cell",
        "my new cell type",
        "hepatocyte",
        "hematopoietic stem cell",
        "B cell",
    }
    assert set(artifact.labels.get(features.cell_type, flat_names=True)) == {
        "T cell",
        "my new cell type",
        "hepatocyte",
        "hematopoietic stem cell",
        "B cell",
    }
    assert set(artifact.labels.get(features.cell_type_by_expert).to_list("name")) == {
        "T cell",
        "my new cell type",
        "hepatocyte",
        "hematopoietic stem cell",
        "B cell",
    }
    assert experiment_1 in artifact.ulabels.all()

    # call describe
    artifact.describe()

    # clean up
    artifact.delete(permanent=True)
    bt.Gene.filter().delete()
    bt.Organism.filter().delete()
    ln.Schema.filter().delete()
    ln.Feature.filter().delete()
    bt.CellType.filter().delete()
    bt.Tissue.filter().delete()
    bt.Disease.filter().delete()
    ln.ULabel.filter().delete()


def test_labels_get(get_mini_csv):
    artifact = ln.Artifact(get_mini_csv, description="test")
    # feature doesn't exist
    with pytest.raises(TypeError):
        artifact.labels.get("x")
    # no linked labels
    feature_name_feature = ln.Feature(name="feature name", dtype="cat")
    feature_name_feature.save()
    schema = ln.Schema(features=[feature_name_feature])
    schema.save()
    artifact.save()
    assert str(artifact.features) == "no linked features"
    # test for deprecated add_schema
    artifact.features._add_schema(schema, slot="random")
    assert artifact.feature_sets.first() == schema
    artifact.delete(permanent=True, storage=True)
    schema.delete()
    feature_name_feature.delete()


@pytest.fixture
def get_test_artifacts():
    with open("./default_storage_unit_core/test-inherit1", "w") as f:
        f.write("artifact1")
    with open("./default_storage_unit_core/test-inherit2", "w") as f:
        f.write("artifact2")
    artifact1 = ln.Artifact("./default_storage_unit_core/test-inherit1")
    artifact1.save()
    artifact2 = ln.Artifact("./default_storage_unit_core/test-inherit2")
    artifact2.save()
    yield artifact1, artifact2
    artifact1.delete(permanent=True, storage=True)
    artifact2.delete(permanent=True, storage=True)


# also see test_feature_manager!
def test_add_from(get_test_artifacts):
    artifact1, artifact2 = get_test_artifacts
    label_names = [f"Project {i}" for i in range(3)]
    ulabels = [ln.ULabel(name=label_name) for label_name in label_names]
    ln.save(ulabels)

    cell_line_names = [f"Cell line {i}" for i in range(3)]
    cell_lines = [bt.CellLine(name=name) for name in cell_line_names]
    ln.save(cell_lines)

    # pass a list of length 0
    artifact2.labels.add([])
    # now actually pass the labels
    artifact2.labels.add(ulabels)
    # here test add without passing a feature
    artifact2.labels.add(cell_lines)
    assert artifact2.cell_lines.count() == len(cell_lines)

    assert artifact1.ulabels.exists() is False
    artifact1.labels.add_from(artifact2)
    assert artifact1.ulabels.count() == artifact2.ulabels.count()
    assert artifact1.cell_lines.count() == artifact2.cell_lines.count()

    artifact2.cell_lines.remove(*cell_lines)
    artifact1.cell_lines.remove(*cell_lines)
    artifact2.ulabels.remove(*ulabels)
    artifact1.ulabels.remove(*ulabels)

    for ulabel in ulabels:
        ulabel.delete()
    for cell_line in cell_lines:
        cell_line.delete()



# Generated by Django 5.2 on 2025-08-09 13:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0118_alter_recordproject_value_projectrecord"),
    ]

    operations = [
        migrations.RenameField(
            model_name="project",
            old_name="records",
            new_name="linked_in_records",
        ),
        migrations.AddField(
            model_name="project",
            name="records",
            field=models.ManyToManyField(
                related_name="projects",
                through="lamindb.ProjectRecord",
                to="lamindb.record",
            ),
        ),
    ]



# Generated by Django 5.1.4 on 2025-04-30 09:11

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0090_runproject_project_runs"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="featurevalue",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="space",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="user",
            options={"base_manager_name": "objects"},
        ),
    ]



import lamindb as ln
import pandas as pd
import pytest
from lamindb import UPath
from lamindb.models._is_versioned import (
    bump_version,
    get_new_path_from_uid,
    set_version,
)


@pytest.fixture(scope="module")
def df1():
    return pd.DataFrame({"feat1": [1, 2]})


@pytest.fixture(scope="module")
def df2():
    return pd.DataFrame({"feat1": [2, 3]})


def test_set_version():
    # all remaining lines are covered in notebooks
    with pytest.raises(ValueError):
        set_version(None, "weird-version")
    assert set_version(None, "1.2") == "2"
    assert set_version(None, "0") == "1"
    assert set_version(None, "1") == "2"
    assert set_version("1.2.3", "0") == "1.2.3"
    assert set_version("1.2.3") == "1.2.3"


def test_bump_version():
    current_version_major_only = "2"
    current_version_major_minor = "2.1"
    weird_version = "weird-version"
    with pytest.raises(ValueError):
        bump_version(weird_version)
    assert bump_version(weird_version, behavior="ignore") == "?"
    assert bump_version(current_version_major_only, bump_type="major") == "3"
    assert bump_version(current_version_major_only, bump_type="minor") == "2.1"
    assert bump_version(current_version_major_minor, bump_type="major") == "3"
    assert bump_version(current_version_major_minor, bump_type="minor") == "2.2"


def test__add_to_version_family(df1, df2):
    artifact1 = ln.Artifact.from_dataframe(df1, description="test1")
    artifact1.save()
    artifact2 = ln.Artifact.from_dataframe(df2, description="test2")
    artifact2.save()
    assert (
        artifact1.uid[: artifact1._len_stem_uid]
        != artifact2.uid[: artifact2._len_stem_uid]
    )
    artifact2._add_to_version_family(artifact1)
    assert (
        artifact1.uid[: artifact1._len_stem_uid]
        == artifact2.uid[: artifact2._len_stem_uid]
    )
    assert (
        artifact1.path.name[: artifact1._len_stem_uid]
        == artifact2.path.name[: artifact2._len_stem_uid]
    )
    artifact1.delete(permanent=True)
    artifact2.delete(permanent=True)


def test_get_new_path_from_uid():
    # test cloud path as it has different behavior than local path
    with open("test_new_path.txt", "w") as f:
        f.write("test_new_path")
    old_path = UPath("s3://lamindata/.lamindb/test_new_path.txt")
    old_path.upload_from("./test_new_path.txt")
    assert old_path.exists()
    new_path = get_new_path_from_uid(
        old_path=old_path,
        old_uid="test_new_path",
        new_uid="test_new_path2",
    )
    assert new_path == "test_new_path2.txt"
    new_path = old_path.rename(new_path)
    assert new_path.exists()
    assert str(new_path) == "s3://lamindata/.lamindb/test_new_path2.txt"
    assert not old_path.exists()
    new_path.unlink()
    UPath("./test_new_path.txt").unlink()


def test_latest_version_and_get():
    # build one version family
    transform_v1 = ln.Transform(description="Introduction").save()
    assert transform_v1.is_latest
    assert transform_v1.version is None
    # pass the latest version, also vary the name for the fun of it
    transform_v2 = ln.Transform(
        description="Introduction v2", revises=transform_v1, version="2"
    ).save()
    assert not transform_v1.is_latest
    assert transform_v2.is_latest
    assert transform_v2.uid.endswith("0001")
    # consciously *not* pass the latest version to revises but the previous
    # it automatically retrieves the latest version
    transform_v3 = ln.Transform(description="Introduction", revises=transform_v1).save()
    assert transform_v3.uid.endswith("0002")
    assert not ln.Transform.objects.get(
        description="Introduction v2", version="2"
    ).is_latest
    assert transform_v3.is_latest
    transform_v4 = ln.Transform(description="Introduction").save()
    assert transform_v4.is_latest
    # add another transform with the same name that's not part of this family
    # but will also be a hit for the query
    assert len(ln.Transform.filter(description="Introduction").all()) == 3
    assert len(ln.Transform.filter(description="Introduction").latest_version()) == 2
    transform_v4.delete(permanent=True)
    assert ln.Transform.get(description="Introduction") == transform_v3
    assert (
        ln.Transform.filter(description="Introduction").latest_version().one()
        == transform_v3
    )

    # test get
    assert ln.Transform.get(transform_v3.uid) == transform_v3
    assert ln.Transform.get(transform_v3.id) == transform_v3
    assert ln.Transform.get(transform_v3.uid[:4]) == transform_v3

    # test delete
    transform_v3.delete(permanent=True)
    assert transform_v2.is_latest

    # test empty QuerySet
    assert (
        ln.Transform.filter(description="IntroductionNotExists")
        .latest_version()
        .one_or_none()
        is None
    )



# Generated by Django 5.2 on 2025-02-13 12:00

import django.db.models.deletion
from django.db import migrations, models

import lamindb.base.fields


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0086_various"),
    ]

    operations = [
        migrations.RenameField(
            model_name="artifact",
            old_name="_schemas_m2m",
            new_name="feature_sets",
        ),
        migrations.AlterField(
            model_name="artifact",
            name="schema",
            field=lamindb.base.fields.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="validated_artifacts",
                to="lamindb.schema",
            ),
        ),
        migrations.AlterField(
            model_name="artifact",
            name="feature_sets",
            field=models.ManyToManyField(
                related_name="artifacts",
                through="lamindb.ArtifactSchema",
                to="lamindb.schema",
            ),
        ),
    ]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal

import h5py
from anndata._io.specs.registry import get_spec

from ._anndata_accessor import AnnDataAccessor, StorageType, registry
from ._polars_lazy_df import POLARS_SUFFIXES, _open_polars_lazy_df
from ._pyarrow_dataset import PYARROW_SUFFIXES, _open_pyarrow_dataset
from ._spatialdata_accessor import SpatialDataAccessor
from ._tiledbsoma import _open_tiledbsoma
from .paths import filepath_from_artifact

if TYPE_CHECKING:
    from collections.abc import Iterator

    from fsspec.core import OpenFile
    from polars import LazyFrame as PolarsLazyFrame
    from pyarrow.dataset import Dataset as PyArrowDataset
    from tiledbsoma import Collection as SOMACollection
    from tiledbsoma import Experiment as SOMAExperiment
    from tiledbsoma import Measurement as SOMAMeasurement
    from upath import UPath

    from lamindb.models.artifact import Artifact


# this dynamically creates a subclass of a context manager class
# and reassigns it to an instance of the superclass
# so that the instance calls finalize on close or exit
def _track_writes_factory(obj: Any, finalize: Callable):
    closed: bool = False

    tracked_class = obj.__class__
    type_dict = {"__doc__": tracked_class.__doc__}
    if hasattr(tracked_class, "__slots__"):
        type_dict["__slots__"] = ()
    if hasattr(tracked_class, "__exit__"):

        def __exit__(self, exc_type, exc_val, exc_tb):
            nonlocal closed
            tracked_class.__exit__(self, exc_type, exc_val, exc_tb)
            if not closed:
                finalize()
                closed = True

        type_dict["__exit__"] = __exit__
    if hasattr(tracked_class, "close"):

        def close(self, *args, **kwargs):
            nonlocal closed
            tracked_class.close(self, *args, **kwargs)
            if not closed:
                finalize()
                closed = True

        type_dict["close"] = close

    Track = type(tracked_class.__name__ + "Track", (tracked_class,), type_dict)
    obj.__class__ = Track
    return obj


@dataclass
class BackedAccessor:
    """h5py.File or zarr.Group accessor."""

    connection: OpenFile
    """The connection."""
    storage: StorageType
    """The storage access."""


def backed_access(
    artifact_or_filepath: Artifact | UPath,
    mode: str = "r",
    engine: Literal["pyarrow", "polars"] = "pyarrow",
    using_key: str | None = None,
    **kwargs,
) -> (
    AnnDataAccessor
    | SpatialDataAccessor
    | BackedAccessor
    | SOMACollection
    | SOMAExperiment
    | SOMAMeasurement
    | PyArrowDataset
    | Iterator[PolarsLazyFrame]
):
    from lamindb.models import Artifact

    if isinstance(artifact_or_filepath, Artifact):
        artifact = artifact_or_filepath
        objectpath, _ = filepath_from_artifact(artifact, using_key=using_key)
    else:
        artifact = None
        objectpath = artifact_or_filepath
    name = objectpath.name
    # ignore .gz, only check the real suffix
    suffixes = objectpath.suffixes
    suffix = (
        suffixes[-2] if len(suffixes) > 1 and ".gz" in suffixes else objectpath.suffix
    )

    if name == "soma" or suffix == ".tiledbsoma":
        if mode not in {"r", "w"}:
            raise ValueError("`mode` should be either 'r' or 'w' for tiledbsoma.")
        return _open_tiledbsoma(objectpath, mode=mode, **kwargs)  # type: ignore
    elif suffix in {".h5", ".hdf5", ".h5ad"}:
        conn, storage = registry.open("h5py", objectpath, mode=mode, **kwargs)
    elif suffix == ".zarr":
        if mode not in {"r", "r+"}:
            raise ValueError("`mode` should be either 'r' or 'r+' for zarr.")
        conn, storage = registry.open("zarr", objectpath, mode=mode, **kwargs)
        if "spatialdata_attrs" in storage.attrs:
            return SpatialDataAccessor(storage, name, artifact)
    elif len(df_suffixes := _flat_suffixes(objectpath)) == 1 and (
        df_suffix := df_suffixes.pop()
    ) in set(PYARROW_SUFFIXES).union(POLARS_SUFFIXES):
        return _open_dataframe(objectpath, df_suffix, engine, **kwargs)
    else:
        raise ValueError(
            "The object should have .h5, .hdf5, .h5ad, .zarr, .tiledbsoma suffix "
            f"be compatible with pyarrow.dataset.dataset or polars.scan_* functions, "
            f"instead of being {suffix} object."
        )

    is_anndata = suffix == ".h5ad" or get_spec(storage).encoding_type == "anndata"
    if is_anndata:
        if mode != "r" and isinstance(storage, h5py.Group):
            raise ValueError("Can only access `hdf5` `AnnData` with mode='r'.")
        return AnnDataAccessor(conn, storage, name, artifact)
    else:
        return BackedAccessor(conn, storage)


def _flat_suffixes(paths: UPath | list[UPath]) -> set[str]:
    # it is assumed here that the paths exist
    # we don't check here that the filesystem is the same
    # but this is a requirement for pyarrow.dataset.dataset
    path_list = []
    if isinstance(paths, Path):
        paths = [paths]
    for path in paths:
        # assume http is always a file
        if getattr(path, "protocol", None) not in {"http", "https"} and path.is_dir():
            path_list += [p for p in path.rglob("*") if p.suffix != ""]
        else:
            path_list.append(path)

    suffixes = set()
    for path in path_list:
        path_suffixes = path.suffixes
        # this doesn't work for externally gzipped files, REMOVE LATER
        path_suffix = (
            path_suffixes[-2]
            if len(path_suffixes) > 1 and ".gz" in path_suffixes
            else path.suffix
        )
        suffixes.add(path_suffix)
    return suffixes


def _open_dataframe(
    paths: UPath | list[UPath],
    suffix: str | None = None,
    engine: Literal["pyarrow", "polars"] = "pyarrow",
    **kwargs,
) -> PyArrowDataset | Iterator[PolarsLazyFrame]:
    if engine not in {"pyarrow", "polars"}:
        raise ValueError(
            f"Unknown engine: {engine}. It should be 'pyarrow' or 'polars'."
        )

    df_suffix: str
    if suffix is None:
        df_suffixes = _flat_suffixes(paths)
        if len(df_suffixes) > 1:
            raise ValueError(
                f"The artifacts in the collection have different file formats: {', '.join(df_suffixes)}.\n"
                "It is not possible to open such stores with pyarrow or polars."
            )
        df_suffix = df_suffixes.pop()
    else:
        df_suffix = suffix

    if engine == "pyarrow" and df_suffix not in PYARROW_SUFFIXES:
        raise ValueError(
            f"{df_suffix} files are not supported by pyarrow, "
            f"they should have one of these formats: {', '.join(PYARROW_SUFFIXES)}."
        )
    elif engine == "polars" and df_suffix not in POLARS_SUFFIXES:
        raise ValueError(
            f"{df_suffix} files are not supported by polars, "
            f"they should have one of these formats: {', '.join(POLARS_SUFFIXES)}."
        )

    polars_without_fsspec = engine == "polars" and not kwargs.get("use_fsspec", False)
    if (engine == "pyarrow" or polars_without_fsspec) and not isinstance(paths, Path):
        # this checks that the filesystem is the same for all paths
        # this is a requirement of pyarrow.dataset.dataset
        fs = getattr(paths[0], "fs", None)
        for path in paths[1:]:
            # this assumes that the filesystems are cached by fsspec
            if getattr(path, "fs", None) is not fs:
                engine_msg = (
                    "polars engine without passing `use_fsspec=True`"
                    if engine == "polars"
                    else "pyarrow engine"
                )
                raise ValueError(
                    "The collection has artifacts with different filesystems, "
                    f"this is not supported for {engine_msg}."
                )

    return (
        _open_pyarrow_dataset(paths, **kwargs)
        if engine == "pyarrow"
        else _open_polars_lazy_df(paths, **kwargs)
    )



import subprocess
import sys
import time
from pathlib import Path

import lamindb as ln
import lamindb_setup as ln_setup
import pytest
from lamindb._finish import clean_r_notebook_html, get_shortcut
from lamindb.core._context import LogStreamTracker, context
from lamindb.errors import TrackNotCalled, ValidationError

SCRIPTS_DIR = Path(__file__).parent.resolve() / "scripts"
NOTEBOOKS_DIR = Path(__file__).parent.resolve() / "notebooks"


def test_track_basic_invocation():
    project = "non-existing project"
    with pytest.raises(ln.errors.InvalidArgument) as error:
        ln.track(project=project)
    assert (
        error.exconly()
        == f"lamindb.errors.InvalidArgument: Project '{project}' not found, either create it with `ln.Project(name='...').save()` or fix typos."
    )
    space = "non-existing space"
    with pytest.raises(ln.errors.InvalidArgument) as error:
        ln.track(space=space)
    assert (
        error.exconly()
        == f"lamindb.errors.InvalidArgument: Space '{space}', please check on the hub UI whether you have the correct `uid` or `name`."
    )

    predecessor1 = ln.Transform(key="parent 1").save()
    predecessor2 = ln.Transform(key="parent 2").save()
    successor = ln.Transform(key="successor").save()
    successor.predecessors.set([predecessor1, predecessor2])

    # first invocation
    params = {"param1": 1, "param2": "my-string", "param3": 3.14}
    with pytest.raises(ValidationError) as exc:
        ln.track(transform=successor, params=params)
    assert (
        exc.exconly()
        == """lamindb.errors.ValidationError: These keys could not be validated: ['param1', 'param2', 'param3']
Here is how to create a feature:

  ln.Feature(name='param1', dtype='int').save()
  ln.Feature(name='param2', dtype='cat ? str').save()
  ln.Feature(name='param3', dtype='float').save()"""
    )
    ln.Feature(name="param1", dtype="int").save()
    ln.Feature(name="param2", dtype="str").save()
    ln.Feature(name="param3", dtype="float").save()
    ln.track(transform=successor, params=params)
    print("outside", id(ln.context))
    assert ln.context.run.features.get_values() == params
    # second invocation
    params = {"param1": 1, "param2": "my-string", "param3": 3.14, "param4": [1, 2]}
    param4 = ln.Feature(name="param4", dtype="int").save()
    with pytest.raises(ValidationError) as exc:
        ln.track(transform=successor, params=params)
    assert (
        exc.exconly()
        == """lamindb.errors.ValidationError: Expected dtype for 'param4' is 'int', got 'list[int]'"""
    )
    # fix param4 dtype
    param4.dtype = "list[int]"
    param4.save()
    # re-run
    ln.track(transform=successor, params=params)
    assert ln.context.run.features.get_values() == params

    # test that run populates things like ULabels etc.
    ulabel = ln.ULabel(name="my-label-in-track")
    assert ulabel.run == ln.context.run

    # test that we can call ln.finish() also for pipeline-like transforms
    assert ln.context.run.finished_at is None
    ln.finish()
    assert ln.context.run.finished_at is not None

    # unset to remove side effects
    ln.context._run = None


def test_track_notebook_colab():
    notebook_path = "/fileId=1KskciVXleoTeS_OGoJasXZJreDU9La_l"
    ln.context._track_notebook(path_str=notebook_path)


def test_finish_before_track():
    ln.context._run = None
    with pytest.raises(TrackNotCalled) as error:
        ln.finish()
    assert "Please run `ln.track()` before `ln.finish()" in error.exconly()


def test_invalid_transform_type():
    transform = ln.Transform(key="test transform")
    ln.track(transform=transform)
    ln.context._path = None
    ln.context.run.transform.type = "script"
    with pytest.raises(ValueError) as error:
        ln.finish()
    assert "Transform type is not allowed to be" in error.exconly()

    # unset to remove side effects
    ln.context._run = None


def test_create_or_load_transform():
    title = "title"
    version = "2.0"
    uid = "NJvdsWWbJlZS0000"
    context.uid = uid
    context.version = version
    context._path = Path("my-test-transform-create-or-load.py")
    context._path.touch(exist_ok=True)
    context._create_or_load_transform(
        description=title,
        transform_type="notebook",
    )
    assert context._transform.uid == uid
    assert context._transform.version == version
    assert context._transform.description == title
    context._create_or_load_transform(
        description=title,
    )
    assert context._transform.uid == uid
    assert context._transform.version == version
    assert context._transform.description == title

    # now, test an updated transform name
    context._create_or_load_transform(
        description="updated title",
    )
    assert context._transform.uid == uid
    assert context._transform.version == version
    assert context._transform.description == "updated title"

    # unset to remove side effects
    ln.context._uid = None
    ln.context._run = None
    ln.context._transform = None
    ln.context._path.unlink()
    ln.context._path = None


def test_run_scripts():
    # regular execution
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'script-to-test-versioning.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 0
    assert (
        "created Transform('Ro1gl7n8YrdH0000'), started new Run("
        in result.stdout.decode()
    )

    # updated key (filename change)
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'script-to-test-filename-change.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 0
    assert "renaming transform" in result.stdout.decode()

    # version already taken
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'duplicate1/script-to-test-versioning.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 1
    assert (
        "✗ version '1' is already taken by Transform('Ro1gl7n8YrdH0000'); please set another version, e.g., ln.context.version = '1.1'"
        in result.stderr.decode()
    )

    # regular version bump
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'duplicate2/script-to-test-versioning.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 0
    assert (
        "created Transform('Ro1gl7n8YrdH0002'), started new Run("
        in result.stdout.decode()
    )
    assert not ln.Transform.get("Ro1gl7n8YrdH0001").is_latest
    assert ln.Transform.get("Ro1gl7n8YrdH0002").is_latest

    # inconsistent version
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'duplicate3/script-to-test-versioning.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 1
    assert (
        "Transform is already tagged with version 2, but you passed 3"
        in result.stderr.decode()
    )

    # multiple folders, do not match the key because of the folder structure
    ln.Transform.filter(key__endswith="script-to-test-versioning.py").update(
        key="teamA/script-to-test-versioning.py"
    )
    # this test creates a transform with key script-to-test-versioning.py at the root level
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'duplicate4/script-to-test-versioning.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 0
    assert "ignoring transform" in result.stdout.decode()

    transform = ln.Transform.get(key="script-to-test-versioning.py")

    # multiple folders, match the key, also test is finished
    result = subprocess.run(  # noqa: S602
        f"python {SCRIPTS_DIR / 'duplicate5/script-to-test-versioning.py'}",
        shell=True,
        capture_output=True,
    )
    assert result.returncode == 0
    assert f"{transform.stem_uid}" in result.stdout.decode()
    assert "making new version" in result.stdout.decode()

    transform = ln.Transform.get(key="script-to-test-versioning.py")
    assert transform.latest_run.finished_at is not None


def test_run_external_script():
    script_path = "sub/lamin-cli/tests/scripts/run-track-and-finish-sync-git.py"
    result = subprocess.run(  # noqa: S602
        f"python {script_path}",
        shell=True,
        capture_output=True,
    )
    print(result.stdout.decode())
    print(result.stderr.decode())
    assert result.returncode == 0
    assert "created Transform" in result.stdout.decode()
    assert "started new Run" in result.stdout.decode()
    transform = ln.Transform.get(key="run-track-and-finish-sync-git.py")
    # the algorithm currently picks different commits depending on the state of the repo
    # any of these commits are valid
    assert transform.uid == "m5uCHTTpJnjQ0000"
    assert transform.reference.endswith(
        "/tests/scripts/run-track-and-finish-sync-git.py"
    )
    assert transform.reference.startswith(
        "https://github.com/laminlabs/lamin-cli/blob/"
    )
    assert transform.reference_type == "url"
    assert transform.description == "My good script"
    # ensure that the source code is not saved as an output artifact
    assert transform.latest_run.output_artifacts.count() == 0
    assert transform.runs.count() == 1
    assert transform.hash == "VC1oTPcaVSrzNrXUT9p4qw"


@pytest.mark.parametrize("type", ["notebook", "script"])
def test_track_notebook_or_script_manually(type):
    transform = ln.Transform(key="My notebook", type=type)
    with pytest.raises(ValueError) as error:
        ln.track(transform=transform)
    assert (
        error.exconly()
        == "ValueError: Use `ln.track()` without passing transform in a notebook or script - metadata is automatically parsed"
    )


def test_clean_r_notebook_html():
    orig_notebook_path = NOTEBOOKS_DIR / "basic-r-notebook.Rmd.html"
    content = orig_notebook_path.read_text()
    orig_notebook_path.write_text(content.replace("SHORTCUT", get_shortcut()))
    comparison_path = NOTEBOOKS_DIR / "basic-r-notebook.Rmd.cleaned.html"
    compare = comparison_path.read_text()
    comparison_path.unlink()
    title_text, cleaned_path = clean_r_notebook_html(orig_notebook_path)
    assert comparison_path == cleaned_path
    assert title_text == "My exemplary R analysis"
    assert compare == cleaned_path.read_text()  # check that things have been stripped
    comparison_path.write_text(compare)
    orig_notebook_path.write_text(content.replace(get_shortcut(), "SHORTCUT"))


class MockRun:
    def __init__(self, uid):
        self.uid = uid
        self.report = None
        self.saved = False

    def save(self):
        self.saved = True


def test_logstream_tracker_multiple():
    tracker1 = LogStreamTracker()
    tracker2 = LogStreamTracker()
    tracker3 = LogStreamTracker()

    try:
        # Start trackers one by one and print messages
        print("Initial stdout")

        tracker1.start(MockRun("run1"))
        print("After starting tracker1")

        tracker2.start(MockRun("run2"))
        print("After starting tracker2")

        tracker3.start(MockRun("run3"))
        print("After starting tracker3")

        print("Testing stderr", file=sys.stderr)

        time.sleep(0.1)

        # Clean up in reverse order
        tracker3.finish()
        tracker2.finish()
        tracker1.finish()

        # Verify log contents - each log should only contain messages after its start
        expected_contents = {
            1: [
                "After starting tracker1",
                "After starting tracker2",
                "After starting tracker3",
                "Testing stderr",
            ],
            2: ["After starting tracker2", "After starting tracker3", "Testing stderr"],
            3: ["After starting tracker3", "Testing stderr"],
        }

        for i in range(1, 4):
            log_path = Path(ln_setup.settings.cache_dir / f"run_logs_run{i}.txt")
            with open(log_path) as f:
                content = f.read()
                print(f"\nContents of run{i} log:")
                print(content)
                # Check each expected line is in the content
                for expected_line in expected_contents[i]:
                    assert expected_line in content, (
                        f"Expected '{expected_line}' in log {i}"
                    )

                # Check earlier messages are NOT in the content
                if i > 1:
                    assert "Initial stdout" not in content
                    assert "After starting tracker" + str(i - 1) not in content

    finally:
        # Cleanup
        for i in range(1, 4):
            log_path = Path(ln_setup.settings.cache_dir / f"run_logs_run{i}.txt")
            if log_path.exists():
                log_path.unlink()


def test_logstream_tracker_exception_handling():
    tracker = LogStreamTracker()
    original_excepthook = sys.excepthook
    run = MockRun("error")

    try:
        tracker.start(run)
        print("Before error")

        # Create and capture exception info
        exc_type = ValueError
        exc_value = ValueError("Test error")
        exc_traceback = None
        try:
            raise exc_value
        except ValueError:
            exc_traceback = sys.exc_info()[2]

        # Handle the exception - this will trigger cleanup
        tracker.handle_exception(exc_type, exc_value, exc_traceback)

        # Verify run status
        assert run.saved
        assert run.report is not None

        # Verify the content was written before cleanup
        content = run.report.cache().read_text()
        print("Log contents:", content)
        assert "Before error" in content
        assert "ValueError: Test error" in content
        assert "Traceback" in content

    finally:
        sys.excepthook = original_excepthook



# .latest_version is tested in test_versioning.py


import re

import bionty as bt
import lamindb as ln
import pytest
from django.core.exceptions import FieldError
from lamindb.base.users import current_user_id
from lamindb.errors import InvalidArgument
from lamindb.models.query_set import DoesNotExist


# please also see the test_curate_df.py tests
def test_df():
    project_label = ln.ULabel(name="project").save()
    project_names = [f"Project {i}" for i in range(3)]
    labels = ln.ULabel.from_values(project_names, create=True).save()
    project_label.children.add(*labels)
    df = ln.ULabel.to_dataframe(include="parents__name")
    assert df.columns[2] == "parents__name"
    assert df["parents__name"].iloc[0] == {project_label.name}
    df = ln.ULabel.to_dataframe(include=["parents__name", "parents__created_by_id"])
    assert df.columns[3] == "parents__created_by_id"
    assert df["parents__name"].iloc[0] == {project_label.name}
    assert set(df["parents__created_by_id"].iloc[0]) == {current_user_id()}

    # for other models
    feature_names = [f"Feature {i}" for i in range(3)]
    features = [ln.Feature(name=name, dtype=int) for name in feature_names]
    ln.save(features)
    schema = ln.Schema(features, name="my schema").save()
    schema.features.set(features)

    df = ln.Schema.filter(name="my schema").to_dataframe(include="features__name")
    assert df.columns[2] == "features__name"
    # order is not conserved
    assert set(df["features__name"].iloc[0]) == set(feature_names)
    # pass a list
    df = ln.Schema.filter(name="my schema").to_dataframe(
        include=["features__name", "features__created_by_id"]
    )
    assert df.columns[3] == "features__created_by_id"
    assert set(df["features__name"].iloc[0]) == set(feature_names)
    assert set(df["features__created_by_id"].iloc[0]) == {current_user_id()}

    # inner join parents on features
    df = ln.Schema.filter().to_dataframe(
        include=["features__name", "features__created_by_id"]
    )
    assert set(df["features__name"].iloc[0]) == set(feature_names)
    assert set(df["features__created_by_id"].iloc[0]) == {current_user_id()}

    # raise error for non many-to-many
    df = ln.ULabel.filter(name="Project 0").to_dataframe(include="created_by__name")
    assert df["created_by__name"].iloc[0] == "Test User1"

    # do not return fields with no data in the registry
    # does not make sense in Alex's opinion
    # too much magic; got removed in https://github.com/laminlabs/lamindb/pull/2238
    # df = (
    #     ln.Artifact.using("laminlabs/cellxgene")
    #     .filter(suffix=".h5ad")
    #     .to_dataframe(include=["tissues__name", "pathways__name"])
    # )
    # assert "tissues__name" in df.columns
    # assert "pathways__name" not in df.columns
    # assert df.shape[0] > 0

    # clean up
    project_label.delete(permanent=True)
    for label in labels:
        label.delete(permanent=True)

    schema.delete(permanent=True)
    for feature in features:
        feature.delete(permanent=True)

    # call it from a non-select-derived queryset
    qs = ln.User.objects.all()
    assert qs.to_dataframe().iloc[0]["handle"] == "testuser1"


def test_one_first():
    qs = ln.User.objects.all()
    assert qs.one().handle == "testuser1"
    assert qs.first().handle == "testuser1"
    assert qs.one_or_none().handle == "testuser1"

    qs = ln.User.filter(handle="test")
    with pytest.raises(DoesNotExist):
        qs.one()
    qs = bt.Source.filter().all()
    # should be MultipleResultsFound, but internal to Django
    with pytest.raises(Exception):  # noqa: B017
        qs.one()
    with pytest.raises(Exception):  # noqa: B017
        qs.one_or_none()


def test_filter_related_field_name():
    with pytest.raises(
        FieldError,
        match=re.escape(
            "Invalid lookup 'somelabel' for ulabels. Did you mean ulabels__name?"
        ),
    ):
        ln.Artifact.filter(ulabels="somelabel").all()


def test_filter_unknown_field():
    with pytest.raises(InvalidArgument) as error:
        ln.Artifact.filter(nonexistent="value").all()
    assert error.exconly() == (
        "lamindb.errors.InvalidArgument: You can query either by available fields: branch, cell_lines, cell_markers, cell_types, collections, created_at, created_by, description, developmental_stages, diseases, ethnicities, experimental_factors, feature_sets, genes, hash, id, input_of_runs, is_latest, key, kind, linked_in_records, n_files, n_observations, organisms, otype, pathways, phenotypes, projects, proteins, records, references, run, schema, size, space, storage, suffix, tissues, transform, uid, ulabels, updated_at, version, visibility\n"
        "Or fix invalid feature names: nonexistent"
    )


def test_get_id_type_error():
    with pytest.raises(
        ValueError, match=re.escape("Field 'id' expected a number but got 'abc'.")
    ):
        ln.Artifact.get(id="abc")


def test_get_related_field_name():
    with pytest.raises(
        FieldError,
        match=re.escape(
            "Invalid lookup 'somelabel' for ulabels. Did you mean ulabels__name?"
        ),
    ):
        ln.Artifact.get(ulabels="somelabel").all()


def test_get_unknown_field():
    with pytest.raises(
        FieldError,
        match=re.escape(
            "Unknown field 'nonexistent'. Available fields: branch, cell_lines, cell_markers, cell_types, collections, created_at, created_by, description, developmental_stages, diseases, ethnicities, experimental_factors, feature_sets, genes, hash, id, input_of_runs, is_latest, key, kind, linked_in_records, n_files, n_observations, organisms, otype, pathways, phenotypes, projects, proteins, records, references, run, schema, size, space, storage, suffix, tissues, transform, uid, ulabels, updated_at, version, visibility"
        ),
    ):
        ln.Artifact.get(nonexistent="value")


def test_search():
    label_names = [f"ULabel {i}" for i in range(3)]
    labels = [ln.ULabel(name=name) for name in label_names]
    ln.save(labels)
    qs = ln.ULabel.filter(name__startswith="ULabel").all()
    assert qs.search("ULabel 1")[0].name == "ULabel 1"
    assert qs.search("ULabel 1", field=ln.ULabel.name)[0].name == "ULabel 1"
    for label in labels:
        label.delete(permanent=True)


def test_lookup():
    qs = ln.User.filter(handle="testuser1").all()
    # pass str to field
    lookup = qs.lookup(field="handle")
    assert lookup.testuser1.handle == "testuser1"
    # pass StrField to field
    lookup = qs.lookup(field=ln.User.handle)
    assert lookup.testuser1.handle == "testuser1"
    # manager, default field
    qsm = ln.User.filter(handle="testuser1")
    lookup = qsm.lookup()
    assert lookup.testuser1.handle == "testuser1"


def test_inspect():
    qs = ln.User.filter(handle="testuser1").all()
    assert qs.inspect(["user1", "user2"], "name")["validated"] == []
    assert ln.User.inspect(["user1", "user2"], "name")["validated"] == []
    assert ln.User.inspect(["user1", "user2"], ln.User.name)["validated"] == []
    assert ln.User.inspect("user1", "name")["validated"] == []


def test_validate():
    qs = ln.User.filter(handle="testuser1").all()
    assert qs.validate(["testuser1", "Test User1"], "handle").tolist() == [True, False]
    assert ln.User.validate(["testuser1", "Test User1"], "handle").tolist() == [
        True,
        False,
    ]
    assert ln.User.validate(["testuser1", "Test User1"], ln.User.handle).tolist() == [
        True,
        False,
    ]
    # returns True
    assert ln.User.validate("testuser1", ln.User.handle)


def test_map_synonyms():
    qs = ln.User.filter(handle="testuser1").all()
    assert qs.standardize(["user1", "user2"]) == ["user1", "user2"]


def test_get_doesnotexist_error():
    non_existent_label = "some-label-name"

    with pytest.raises(DoesNotExist) as excinfo:
        ln.ULabel.get(non_existent_label)

    error_message = str(excinfo.value)
    assert f"No record found with uid '{non_existent_label}'" in error_message
    assert (
        f"Did you forget a keyword as in ULabel.get(name='{non_existent_label}')?"
        in error_message
    )



class CreationSettings:
    search_names: bool = True
    """Switch off to speed up creating records (default `True`).

    If `True`, search for alternative names and avoids duplicates.

    FAQ: :doc:`/faq/idempotency`
    """
    artifact_skip_size_hash: bool = False
    """To speed up registering high numbers of files (default `False`).

    This bypasses queries for size and hash to AWS & GCP.

    It speeds up file creation by about a factor 100.
    """
    artifact_silence_missing_run_warning: bool = False
    """Silence warning about missing run & transform during artifact creation (default `False`)."""
    _artifact_use_virtual_keys: bool = True
    """Treat `key` parameter in :class:`~lamindb.Artifact` as virtual.

    If `True`, the `key` is **not** used to construct file paths, but file paths are
    based on the `uid` of artifact.
    """


creation_settings = CreationSettings()



import lamindb as ln
import bionty as bt


# define the global obs schema
obs_schema = ln.Schema(
    name="mudata_papalexi21_subset_obs_schema",
    features=[
        ln.Feature(name="perturbation", dtype="cat[ULabel[Perturbation]]").save(),
        ln.Feature(name="replicate", dtype="cat[ULabel[Replicate]]").save(),
    ],
).save()

# define the ['rna'].obs schema
obs_schema_rna = ln.Schema(
    name="mudata_papalexi21_subset_rna_obs_schema",
    features=[
        ln.Feature(name="nCount_RNA", dtype=int).save(),
        ln.Feature(name="nFeature_RNA", dtype=int).save(),
        ln.Feature(name="percent.mito", dtype=float).save(),
    ],
).save()

# define the ['hto'].obs schema
obs_schema_hto = ln.Schema(
    name="mudata_papalexi21_subset_hto_obs_schema",
    features=[
        ln.Feature(name="nCount_HTO", dtype=int).save(),
        ln.Feature(name="nFeature_HTO", dtype=int).save(),
        ln.Feature(name="technique", dtype=bt.ExperimentalFactor).save(),
    ],
).save()

# define ['rna'].var schema
var_schema_rna = ln.Schema(
    name="mudata_papalexi21_subset_rna_var_schema",
    itype=bt.Gene.symbol,
    dtype=float,
).save()

# define composite schema
mudata_schema = ln.Schema(
    name="mudata_papalexi21_subset_mudata_schema",
    otype="MuData",
    slots={
        "obs": obs_schema,
        "rna:obs": obs_schema_rna,
        "hto:obs": obs_schema_hto,
        "rna:var": var_schema_rna,
    },
).save()

# curate a MuData
mdata = ln.core.datasets.mudata_papalexi21_subset()
bt.settings.organism = "human"  # set the organism to map gene symbols
curator = ln.curators.MuDataCurator(mdata, mudata_schema)
artifact = curator.save_artifact(key="examples/mudata_papalexi21_subset.h5mu")
assert artifact.schema == mudata_schema



from __future__ import annotations

import inspect
from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Callable, Literal, Union

import h5py
import numpy as np
import pandas as pd
from anndata import AnnData
from anndata import __version__ as anndata_version
from anndata._core.index import _normalize_indices
from anndata._core.views import _resolve_idx
from anndata._io.h5ad import read_dataframe_legacy as read_dataframe_legacy_h5
from anndata._io.specs.registry import (
    get_spec,
    read_elem,
    read_elem_partial,
    write_elem,
)
from anndata.compat import _read_attr
from fsspec.implementations.local import LocalFileSystem
from fsspec.utils import infer_compression
from lamin_utils import logger
from lamindb_setup.core.upath import S3FSMap, infer_filesystem
from packaging import version
from upath import UPath

if TYPE_CHECKING:
    from collections.abc import Mapping

    from fsspec.core import OpenFile
    from lamindb_setup.types import UPathStr

    from lamindb import Artifact


anndata_version_parse = version.parse(anndata_version)

if anndata_version_parse < version.parse("0.9.0"):
    from anndata._core.index import Index
else:
    from anndata.compat import Index

if anndata_version_parse < version.parse("0.10.0"):
    if anndata_version_parse < version.parse("0.9.1"):
        logger.warning(
            "Full backed capabilities are not available for this version of anndata,"
            " please install anndata>=0.9.1."
        )

    from anndata._core.sparse_dataset import SparseDataset

    # try csr for groups with no encoding_type
    class CSRDataset(SparseDataset):
        @property
        def format_str(self) -> str:
            return "csr"

    def sparse_dataset(group):
        return SparseDataset(group)

else:
    if anndata_version_parse >= version.parse("0.11.0"):
        from anndata._core.sparse_dataset import (  # type: ignore
            _CSRDataset as CSRDataset,
        )
    else:
        from anndata._core.sparse_dataset import CSRDataset  # type: ignore
    from anndata._core.sparse_dataset import (
        BaseCompressedSparseDataset as SparseDataset,
    )
    from anndata._core.sparse_dataset import sparse_dataset  # type: ignore

    def _check_group_format(*args):
        pass

    CSRDataset._check_group_format = _check_group_format


# zarr and CSRDataset have problems with full selection
def _subset_sparse(sparse_ds: CSRDataset | SparseDataset, indices):
    has_arrays = isinstance(indices[0], np.ndarray) or isinstance(
        indices[1], np.ndarray
    )
    if not has_arrays and indices == (slice(None), slice(None)):
        return sparse_ds.to_memory()
    else:
        return sparse_ds[indices]


def get_module_name(obj):
    return inspect.getmodule(obj).__name__.partition(".")[0]


def _records_to_df(obj):
    if isinstance(obj, pd.DataFrame):
        return obj

    if hasattr(obj, "dtype") and obj.dtype.names is not None:
        formats = []
        for name, (dt, _) in obj.dtype.fields.items():
            if dt.char == "S":
                new_dt = str(dt).replace("S", "U")
            else:
                new_dt = dt
            formats.append((name, new_dt))
        df = pd.DataFrame(obj.astype(formats, copy=False))
        for index_name in ("index", "_index"):
            if index_name in df.columns:
                return df.set_index(index_name)
            return df
    else:
        return obj


class AccessRegistry:
    def __init__(self):
        self._registry = {}
        self._openers = {}

    def register_open(self, module: str):
        def wrapper(func: Callable):
            self._openers[module] = func
            return func

        return wrapper

    def open(self, module: str, *args, **kwargs):
        if module in self._openers:
            return self._openers[module](*args, **kwargs)
        else:
            raise ValueError(f"Module {module} not found, please install it.")

    def register(self, module: str):
        def wrapper(func: Callable):
            func_name = func.__name__
            if func_name not in self._registry:
                self._registry[func_name] = {}
            self._registry[func_name][module] = func
            return func

        return wrapper

    def __getattr__(self, func_name: str):
        def wrapper(*args, **kwargs):
            func_registry = self._registry[func_name]
            for arg in chain(args, kwargs.values()):
                arg_module = get_module_name(arg)
                if arg_module in func_registry:
                    return func_registry[arg_module](*args, **kwargs)
            raise ValueError(f"{func_name} is not registered for this module.")

        return wrapper


# storage specific functions should be registered and called through the registry
registry = AccessRegistry()


@registry.register_open("h5py")
def open(filepath: UPathStr, mode: str = "r", compression: str | None = "infer"):
    fs, file_path_str = infer_filesystem(filepath)
    # we don't open compressed files directly because we need fsspec to uncompress on .open
    compression = (
        infer_compression(file_path_str) if compression == "infer" else compression
    )
    if isinstance(fs, LocalFileSystem) and compression is None:
        assert mode in {"r", "r+", "a", "w", "w-"}, f"Unknown mode {mode}!"  #  noqa: S101
        return None, h5py.File(file_path_str, mode=mode)
    if mode == "r":
        conn_mode = "rb"
    elif mode == "w":
        conn_mode = "wb"
    elif mode == "a":
        conn_mode = "ab"
    else:
        raise ValueError(f"Unknown mode {mode}! Should be 'r', 'w' or 'a'.")
    conn = fs.open(file_path_str, mode=conn_mode, compression=compression)
    try:
        storage = h5py.File(conn, mode=mode)
    except Exception as e:
        conn.close()
        raise e
    return conn, storage


@registry.register("h5py")
def read_dataframe(elem: h5py.Dataset | h5py.Group):
    if isinstance(elem, h5py.Dataset):
        return read_dataframe_legacy_h5(elem)
    else:
        return read_elem(elem)


@registry.register("h5py")
def safer_read_partial(elem, indices):
    is_dataset = isinstance(elem, h5py.Dataset)
    indices_inverse: list | None = None
    encoding_type = get_spec(elem).encoding_type
    # h5py selection for datasets requires sorted indices
    if is_dataset or encoding_type == "dataframe":
        indices_increasing = []
        indices_inverse = []
        for indices_dim in indices:
            # should be integer or bool
            # ignore bool or increasing unique integers
            if (
                isinstance(indices_dim, np.ndarray)
                and indices_dim.dtype != "bool"
                and not np.all(np.diff(indices_dim) > 0)
            ):
                idx_unique, idx_inverse = np.unique(indices_dim, return_inverse=True)
                indices_increasing.append(idx_unique)
                indices_inverse.append(idx_inverse)
            else:
                indices_increasing.append(indices_dim)
                indices_inverse.append(None)
        indices = tuple(indices_increasing)
        if all(idx is None for idx in indices_inverse):
            indices_inverse = None
    result = None
    if encoding_type == "":
        if is_dataset:
            dims = len(elem.shape)
            if dims == 2:
                result = elem[indices]
            elif dims == 1:
                if indices[0] == slice(None):
                    result = elem[indices[1]]
                elif indices[1] == slice(None):
                    result = elem[indices[0]]
        elif isinstance(elem, h5py.Group):
            try:
                ds = CSRDataset(elem)
                result = _subset_sparse(ds, indices)
            except Exception as e:
                logger.debug(
                    f"Encountered an exception while attempting to subset a sparse dataset by indices.\n{e}"
                )
        if result is None:
            raise ValueError(
                "Can not get a subset of the element of type"
                f" {type(elem).__name__} with an empty spec."
            )
    else:
        result = read_elem_partial(elem, indices=indices)
    if indices_inverse is None:
        return result
    else:
        if indices_inverse[0] is None:
            if len(result.shape) == 2:
                return result[:, indices_inverse[1]]
            else:
                return result[indices_inverse[1]]
        elif indices_inverse[1] is None:
            if isinstance(result, pd.DataFrame):
                return result.iloc[indices_inverse[0]]
            else:
                return result[indices_inverse[0]]
        else:
            return result[tuple(indices_inverse)]


@registry.register("h5py")
def keys(storage: h5py.File):
    attrs_keys: dict[str, list] = {}
    for attr in storage.keys():
        if attr == "X":
            continue
        attr_obj = storage[attr]
        if attr in ("obs", "var") and isinstance(attr_obj, h5py.Dataset):
            keys = list(attr_obj.dtype.fields.keys())
        else:
            keys = list(attr_obj.keys())
        if len(keys) > 0:
            attrs_keys[attr] = keys
    return attrs_keys


ArrayTypes = [h5py.Dataset]
GroupTypes = [h5py.Group]
StorageTypes = [h5py.File]


ZARR_INSTALLED = False
try:
    import zarr

    ZARR_INSTALLED = True
except ImportError:
    pass

if ZARR_INSTALLED:
    from anndata._io.zarr import read_dataframe_legacy as read_dataframe_legacy_zarr

    from ._zarr import IS_ZARR_V3, get_zarr_store

    ArrayTypes.append(zarr.Array)
    GroupTypes.append(zarr.Group)
    StorageTypes.append(zarr.Group)

    @registry.register_open("zarr")
    def open(filepath: UPathStr, mode: Literal["r", "r+", "a", "w", "w-"] = "r"):
        assert mode in {"r", "r+", "a", "w", "w-"}, f"Unknown mode {mode}!"  #  noqa: S101

        store = get_zarr_store(filepath)
        kwargs = {}
        if IS_ZARR_V3 and mode != "r":
            # otherwise unable to write
            kwargs["use_consolidated"] = False
        storage = zarr.open(store, mode=mode, **kwargs)
        # zarr v2 re-initializes the mapper
        # we need to put back the correct one
        # S3FSMap is returned from get_zarr_store only for zarr v2
        if isinstance(store, S3FSMap):
            assert not IS_ZARR_V3  # noqa: S101

            storage.store.map = store
        conn = None
        return conn, storage

    @registry.register("zarr")
    def read_dataframe(elem: Union[zarr.Array, zarr.Group]):  # noqa
        if isinstance(elem, zarr.Array):
            return read_dataframe_legacy_zarr(elem)
        else:
            return read_elem(elem)

    @registry.register("zarr")
    def safer_read_partial(elem, indices):
        encoding_type = get_spec(elem).encoding_type
        if encoding_type == "":
            if isinstance(elem, zarr.Array):
                dims = len(elem.shape)
                if dims == 2:
                    return elem.oindex[indices]
                elif dims == 1:
                    if indices[0] == slice(None):
                        return elem.oindex[indices[1]]
                    elif indices[1] == slice(None):
                        return elem.oindex[indices[0]]
            elif isinstance(elem, zarr.Group):
                try:
                    ds = CSRDataset(elem)
                    return _subset_sparse(ds, indices)
                except Exception as e:
                    logger.debug(
                        f"Encountered an exception while attempting to subset a sparse dataset by indices.\n{e}"
                    )
            raise ValueError(
                "Can not get a subset of the element of type"
                f" {type(elem).__name__} with an empty spec."
            )
        else:
            if encoding_type in ("csr_matrix", "csc_matrix"):
                ds = sparse_dataset(elem)
                return _subset_sparse(ds, indices)
            else:
                indices = tuple(
                    idim.tolist()
                    if isinstance(idim, np.ndarray) and idim.dtype == "bool"
                    else idim
                    for idim in indices
                )
                return read_elem_partial(elem, indices=indices)

    # this is needed because accessing zarr.Group.keys() directly is very slow
    @registry.register("zarr")
    def keys(storage: zarr.Group):
        if IS_ZARR_V3:
            paths = storage._sync_iter(storage.store.list())
        else:
            paths = storage.store.keys()

        attrs_keys: dict[str, list] = {}
        obs_var_arrays = []

        prefix = storage.path
        if prefix == "":
            paths_iter = (path for path in paths)
        else:
            prefix += "/"
            paths_iter = (
                path.removeprefix(prefix) for path in paths if path.startswith(prefix)
            )

        for path in paths_iter:
            if path in (".zattrs", ".zgroup"):
                continue
            parts = path.split("/")
            if len(parts) < 2:
                continue
            attr = parts[0]
            key = parts[1]

            if attr == "X":
                continue

            if attr in ("obs", "var"):
                if attr in obs_var_arrays:
                    continue
                if key == ".zarray":
                    attrs_keys.pop(attr, None)
                    obs_var_arrays.append(attr)

            if attr not in attrs_keys:
                attrs_keys[attr] = []

            if key in (".zattrs", ".zgroup", ".zarray"):
                continue
            attr_keys = attrs_keys[attr]
            if key not in attr_keys:
                attr_keys.append(key)

        for attr in obs_var_arrays:
            attrs_keys[attr] = list(storage[attr].dtype.fields.keys())

        return {attr: keys for attr, keys in attrs_keys.items() if len(keys) > 0}


ArrayTypes = tuple(ArrayTypes)  # type: ignore
GroupTypes = tuple(GroupTypes)  # type: ignore
StorageTypes = tuple(StorageTypes)  # type: ignore


ArrayType = Union[ArrayTypes]  # type: ignore
GroupType = Union[GroupTypes]  # type: ignore
StorageType = Union[StorageTypes]  # type: ignore


def _to_memory(elem):
    if isinstance(elem, ArrayTypes):
        return elem[()]
    elif isinstance(elem, SparseDataset):
        return elem.to_memory()
    else:
        return elem


def _try_backed_full(elem):
    # think what to do for compatibility with old var and obs
    if isinstance(elem, ArrayTypes):
        return elem

    if isinstance(elem, GroupTypes):
        encoding_type = get_spec(elem).encoding_type
        if encoding_type in ("csr_matrix", "csc_matrix"):
            return sparse_dataset(elem)
        if "h5sparse_format" in elem.attrs:
            return sparse_dataset(elem)
        if encoding_type == "" and "indptr" in elem:
            return CSRDataset(elem)

    return read_elem(elem)


def _to_index(elem: np.ndarray):
    if elem.dtype in (np.float64, np.int64):
        elem = elem.astype(str)
    return pd.Index(elem)


def _safer_read_index(elem):
    if isinstance(elem, GroupTypes):
        return _to_index(read_elem(elem[_read_attr(elem.attrs, "_index")]))
    elif isinstance(elem, ArrayTypes):
        indices = None
        for index_name in ("index", "_index"):
            if index_name in elem.dtype.names:
                indices = elem[index_name]
                break
        if indices is not None and len(indices) > 0:
            if isinstance(indices[0], bytes):
                indices = np.frompyfunc(lambda x: x.decode("utf-8"), 1, 1)(indices)
            return _to_index(indices)
        else:
            raise ValueError("Indices not found.")
    else:
        raise ValueError(f"Unknown elem type {type(elem)} when reading indices.")


class _MapAccessor:
    def __init__(self, elem, name, indices=None):
        self.elem = elem
        self.indices = indices
        self.name = name

    def __getitem__(self, key):
        if self.indices is None:
            return _try_backed_full(self.elem[key])
        else:
            return registry.safer_read_partial(self.elem[key], indices=self.indices)

    def keys(self):
        return list(self.elem.keys())

    def __repr__(self):
        """Description of the _MapAccessor object."""
        descr = f"Accessor for the AnnData attribute {self.name}"
        descr += f"\n  with keys: {self.keys()}"
        return descr


def _safer_read_df(elem, indices=None):
    if indices is not None:
        obj = registry.safer_read_partial(elem, indices=indices)
        df = _records_to_df(obj)
    else:
        df = registry.read_dataframe(elem)
    if df.index.dtype in (np.float64, np.int64):
        df.index = df.index.astype(str)
    return df


class _AnnDataAttrsMixin:
    storage: StorageType
    _attrs_keys: Mapping[str, list]

    @cached_property
    def obs(self) -> pd.DataFrame | None:
        if "obs" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        return _safer_read_df(
            self.storage["obs"],  # type: ignore
            indices=(indices[0], slice(None)) if indices is not None else None,
        )

    @cached_property
    def var(self) -> pd.DataFrame | None:
        if "var" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        return _safer_read_df(
            self.storage["var"],  # type: ignore
            indices=(indices[1], slice(None)) if indices is not None else None,
        )

    @cached_property
    def uns(self):
        if "uns" not in self._attrs_keys:
            return None
        return read_elem(self.storage["uns"])

    @cached_property
    def X(self):
        indices = getattr(self, "indices", None)
        if indices is not None:
            return registry.safer_read_partial(self.storage["X"], indices=indices)
        else:
            return _try_backed_full(self.storage["X"])

    @cached_property
    def obsm(self):
        if "obsm" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[0], slice(None))
        return _MapAccessor(self.storage["obsm"], "obsm", indices)

    @cached_property
    def varm(self):
        if "varm" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[1], slice(None))
        return _MapAccessor(self.storage["varm"], "varm", indices)

    @cached_property
    def obsp(self):
        if "obsp" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[0], indices[0])
        return _MapAccessor(self.storage["obsp"], "obsp", indices)

    @cached_property
    def varp(self):
        if "varp" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        if indices is not None:
            indices = (indices[1], indices[1])
        return _MapAccessor(self.storage["varp"], "varp", indices)

    @cached_property
    def layers(self):
        if "layers" not in self._attrs_keys:
            return None
        indices = getattr(self, "indices", None)
        return _MapAccessor(self.storage["layers"], "layers", indices)

    @property
    def obs_names(self):
        return self._obs_names

    @property
    def var_names(self):
        return self._var_names

    @cached_property
    def shape(self):
        return len(self._obs_names), len(self._var_names)

    def to_dict(self):
        prepare_adata = {}

        prepare_adata["X"] = _to_memory(self.X)

        if "uns" in self._attrs_keys:
            prepare_adata["uns"] = self.uns

        for attr in ("obs", "var"):
            if attr in self._attrs_keys:
                prepare_adata[attr] = getattr(self, attr)

        for attr in ("obsm", "varm", "obsp", "varp", "layers"):
            if attr in self._attrs_keys:
                prepare_adata[attr] = {}
                get_attr = getattr(self, attr)
                for key in self._attrs_keys[attr]:
                    prepare_adata[attr][key] = _to_memory(get_attr[key])

        if "raw" in self._attrs_keys:
            prepare_adata["raw"] = self.raw.to_dict()

        return prepare_adata

    def to_memory(self):
        adata = AnnData(**self.to_dict())
        return adata


class AnnDataAccessorSubset(_AnnDataAttrsMixin):
    def __init__(self, storage, indices, attrs_keys, obs_names, var_names, ref_shape):
        self.storage = storage
        self.indices = indices

        self._attrs_keys = attrs_keys
        self._obs_names, self._var_names = obs_names, var_names

        self._ref_shape = ref_shape

    def __getitem__(self, index: Index):
        """Access a subset of the underlying AnnData object."""
        oidx, vidx = _normalize_indices(index, self._obs_names, self._var_names)
        new_obs_names, new_var_names = self._obs_names[oidx], self._var_names[vidx]
        if self.indices is not None:
            oidx = _resolve_idx(self.indices[0], oidx, self._ref_shape[0])
            vidx = _resolve_idx(self.indices[1], vidx, self._ref_shape[1])
        return type(self)(
            self.storage,
            (oidx, vidx),
            self._attrs_keys,
            new_obs_names,
            new_var_names,
            self._ref_shape,
        )

    def __repr__(self):
        """Description of the object."""
        n_obs, n_vars = self.shape
        descr = f"{type(self).__name__} object with n_obs × n_vars = {n_obs} × {n_vars}"
        for attr, keys in self._attrs_keys.items():
            descr += f"\n  {attr}: {keys}"
        return descr

    @cached_property
    def raw(self):
        if "raw" not in self._attrs_keys:
            return None
        prepare_indices = None
        if self.indices is not None:
            oidx = self.indices[0]
            if isinstance(oidx, np.ndarray) or oidx != slice(None):
                prepare_indices = oidx, slice(None)
        return AnnDataRawAccessor(
            self.storage["raw"],
            prepare_indices,
            None,
            self._obs_names,
            None,
            self._ref_shape[0],
        )


class AnnDataRawAccessor(AnnDataAccessorSubset):
    def __init__(
        self, storage_raw, indices, attrs_keys, obs_names, var_names, ref_shape
    ):
        var_raw = storage_raw["var"]

        if var_names is None:
            var_names = _safer_read_index(var_raw)

        if isinstance(ref_shape, int):
            ref_shape = ref_shape, len(var_names)
        elif isinstance(ref_shape, tuple) and len(ref_shape) < 2:
            ref_shape = ref_shape[0], len(var_names)

        if attrs_keys is None:
            attrs_keys = {}
            if isinstance(var_raw, ArrayTypes):
                attrs_keys["var"] = list(var_raw.dtype.fields.keys())
            else:
                # for some reason list(var_raw.keys()) is very slow for zarr
                # maybe also directly get keys from the underlying mapper
                attrs_keys["var"] = list(var_raw)
            if "varm" in storage_raw:
                varm_keys_raw = list(storage_raw["varm"])
                if len(varm_keys_raw) > 0:
                    attrs_keys["varm"] = varm_keys_raw

        super().__init__(
            storage_raw, indices, attrs_keys, obs_names, var_names, ref_shape
        )

    @property
    def raw(self):
        raise AttributeError


class AnnDataAccessor(_AnnDataAttrsMixin):
    """Cloud-backed AnnData."""

    def __init__(
        self,
        connection: OpenFile | None,
        storage: StorageType,
        filename: str,
        artifact: Artifact | None = None,
    ):
        self._conn = connection
        self.storage = storage

        self._attrs_keys = registry.keys(self.storage)

        self._name = filename

        self._obs_names = _safer_read_index(self.storage["obs"])  # type: ignore
        self._var_names = _safer_read_index(self.storage["var"])  # type: ignore

        self._artifact = artifact  # save artifact to update in write mode

        self._updated = False  # track updates in r+ mode for zarr

        self._entered = False  # check that the context manager is used
        self._closed = False

    def close(self):
        """Closes the connection."""
        storage = self.storage
        connection = self._conn

        if self._updated and (artifact := self._artifact) is not None:
            from lamindb.models.artifact import Artifact
            from lamindb.models.sqlrecord import init_self_from_db

            # now self._updated can only be True for zarr
            assert ZARR_INSTALLED  # noqa: S101

            store = storage.store
            keys = storage._sync_iter(store.list()) if IS_ZARR_V3 else store.keys()
            # this checks that there consolidated metadata was written before
            # need to update it
            # zmetadata is in spatialdata sometimes for some reason
            if ".zmetadata" in keys or "zmetadata" in keys:
                zarr.consolidate_metadata(store)

            new_version = Artifact(
                artifact.path, revises=artifact, _is_internal_call=True
            ).save()
            # note: sets _state.db = "default"
            init_self_from_db(artifact, new_version)

        if hasattr(storage, "close"):
            storage.close()
        if hasattr(connection, "close"):
            connection.close()
        self._closed = True

    @property
    def closed(self):
        return self._closed

    def __enter__(self):
        self._entered = True

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __getitem__(self, index: Index) -> AnnDataAccessorSubset:
        """Access a subset of the underlying AnnData object."""
        oidx, vidx = _normalize_indices(index, self._obs_names, self._var_names)
        new_obs_names, new_var_names = self._obs_names[oidx], self._var_names[vidx]
        return AnnDataAccessorSubset(
            self.storage,
            (oidx, vidx),
            self._attrs_keys,
            new_obs_names,
            new_var_names,
            self.shape,
        )

    def __repr__(self):
        """Description of the AnnDataAccessor object."""
        n_obs, n_vars = self.shape
        descr = f"AnnDataAccessor object with n_obs × n_vars = {n_obs} × {n_vars}"
        descr += f"\n  constructed for the AnnData object {self._name}"
        for attr, keys in self._attrs_keys.items():
            descr += f"\n    {attr}: {keys}"
        return descr

    @cached_property
    def raw(self):
        if "raw" not in self._attrs_keys:
            return None
        return AnnDataRawAccessor(
            self.storage["raw"], None, None, self._obs_names, None, self.shape[0]
        )

    def add_column(
        self,
        where: Literal["obs", "var"],
        col_name: str,
        col: np.ndarray | pd.Categorical,
    ):
        """Add a new column to .obs or .var of the underlying AnnData object."""
        df_store = self.storage[where]  # type: ignore
        if getattr(df_store, "read_only", True):
            raise ValueError(
                "You can use .add_column(...) only with zarr in a writable mode."
            )
        write_elem(df_store, col_name, col)
        df_store.attrs["column-order"] = df_store.attrs["column-order"] + [col_name]
        # remind only once if this wasn't updated before and not in the context manager
        if not self._updated and not self._entered and self._artifact is not None:
            logger.important(
                "Do not forget to call .close() after you finish "
                f"working with this accessor for {self._name} "
                "to automatically update the corresponding artifact."
            )

        self._updated = True
        # reset the cached property
        # todo: maybe just append the column if the df was already loaded
        self.__dict__.pop(where, None)
        # update the cached columns
        self._attrs_keys[where].append(col_name)


# get the number of observations in an anndata object or file fast and safely
def _anndata_n_observations(object: UPathStr | AnnData) -> int | None:
    if isinstance(object, AnnData):
        return object.n_obs

    try:
        objectpath = UPath(object)
        suffix = objectpath.suffix
        conn_module = {".h5ad": "h5py", ".zarr": "zarr"}.get(suffix, suffix[1:])
        conn, storage = registry.open(conn_module, objectpath, mode="r")
    except Exception as e:
        logger.warning(f"Could not open {object} to read n_observations: {e}")
        return None

    n_observations: int | None = None
    try:
        obs = storage["obs"]
        if isinstance(obs, GroupTypes):  # type: ignore
            if "_index" in obs.attrs:
                elem_key = _read_attr(obs.attrs, "_index")
            else:
                elem_key = next(iter(obs))
            elem = obs[elem_key]
            if isinstance(elem, ArrayTypes):  # type: ignore
                n_observations = elem.shape[0]
            else:
                # assume standard obs group
                n_observations = elem["codes"].shape[0]
        else:
            n_observations = obs.shape[0]
    except Exception as e:
        logger.warning(f"Could not read n_observations from anndata {object}: {e}")
    finally:
        if hasattr(storage, "close"):
            storage.close()
        if hasattr(conn, "close"):
            conn.close()
    return n_observations



# Generated by Django 5.2 on 2025-05-29 12:02

from django.db import migrations


def fix_artifact_kind(apps, schema_editor):
    Artifact = apps.get_model("lamindb", "Artifact")
    Artifact.objects.filter(kind="__lamindb__").update(kind="__lamindb_run__")


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0102_remove_writelog_branch_code_and_more"),
    ]

    operations = [
        migrations.RunPython(fix_artifact_kind),
        migrations.RemoveField(
            model_name="writelog",
            name="migration_state",
        ),
        migrations.RemoveField(
            model_name="writelog",
            name="table",
        ),
        migrations.RemoveField(
            model_name="writelog",
            name="branch",
        ),
        migrations.RemoveField(
            model_name="writelog",
            name="space",
        ),
        migrations.DeleteModel(
            name="WriteLogLock",
        ),
        migrations.DeleteModel(
            name="MigrationState",
        ),
        migrations.DeleteModel(
            name="TableState",
        ),
        migrations.DeleteModel(
            name="WriteLog",
        ),
    ]



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from django.db import connections
from lamin_utils import logger
from rich.table import Column, Table
from rich.text import Text
from rich.tree import Tree

from lamindb.models import CanCurate, Feature
from lamindb.models._from_values import _format_values
from lamindb.models.save import save
from lamindb.models.sqlrecord import (
    REGISTRY_UNIQUE_FIELD,
    get_name_field,
    transfer_fk_to_default_db_bulk,
    transfer_to_default_db,
)

from ._describe import (
    NAME_WIDTH,
    TYPE_WIDTH,
    VALUES_WIDTH,
    describe_header,
    format_rich_tree,
)
from ._django import get_artifact_with_related, get_related_model
from ._relations import dict_related_model_to_related_name

if TYPE_CHECKING:
    from lamindb.models import Artifact, Collection, SQLRecord
    from lamindb.models.query_set import QuerySet

EXCLUDE_LABELS = {"feature_sets"}


def _get_labels(
    obj, links: bool = False, instance: str | None = None
) -> dict[str, QuerySet]:
    """Get all labels associated with an object as a dictionary.

    This is a generic approach that uses django orm.
    """
    if obj.id is None:
        return {}

    labels = {}
    related_models = dict_related_model_to_related_name(
        obj.__class__, links=links, instance=instance
    )

    for _, related_name in related_models.items():
        if related_name not in EXCLUDE_LABELS and not related_name.startswith("_"):
            labels[related_name] = getattr(obj, related_name).all()
    return labels


def _get_labels_postgres(
    self: Artifact | Collection, m2m_data: dict | None = None
) -> dict[str, dict[int, str]]:
    """Get all labels associated with an artifact or collection as a dictionary.

    This is a postgres-specific approach that uses django Subquery.
    """
    if m2m_data is None:
        artifact_meta = get_artifact_with_related(self, include_m2m=True)
        m2m_data = artifact_meta.get("related_data", {}).get("m2m", {})
    return m2m_data


def describe_labels(
    self: Artifact | Collection,
    labels_data: dict | None = None,
    tree: Tree | None = None,
    as_subtree: bool = False,
):
    """Describe labels associated with an artifact or collection."""
    if not self._state.adding and connections[self._state.db].vendor == "postgresql":
        labels_data = _get_labels_postgres(self, labels_data)
    if not labels_data:
        labels_data = _get_labels(self, instance=self._state.db)

    # initialize tree
    if tree is None:
        tree = describe_header(self)
    if not labels_data:
        return tree

    labels_table = Table(
        Column("", style="", no_wrap=True, width=NAME_WIDTH),
        Column("", style="dim", no_wrap=True, width=TYPE_WIDTH),
        Column("", width=VALUES_WIDTH, no_wrap=True),
        show_header=False,
        box=None,
        pad_edge=False,
    )
    for related_name, labels in labels_data.items():
        if not labels or related_name == "feature_sets":
            continue
        if isinstance(labels, dict):  # postgres, labels are a dict[id, name]
            print_values = _format_values(labels.values(), n=10, quotes=False)
        else:  # labels are a QuerySet
            field = get_name_field(labels)
            print_values = _format_values(
                labels.values_list(field, flat=True), n=10, quotes=False
            )
        if print_values:
            related_model = get_related_model(self, related_name)
            type_str = related_model.__get_name_with_module__()
            labels_table.add_row(
                f".{related_name}", Text(type_str, style="dim"), print_values
            )

    labels_header = Text("Labels", style="bold green_yellow")
    if as_subtree:
        if labels_table.rows:
            labels_tree = Tree(labels_header, guide_style="dim")
            labels_tree.add(labels_table)
            return labels_tree
    else:
        if labels_table.rows:
            labels_tree = tree.add(labels_header)
            labels_tree.add(labels_table)
        return tree


def _save_validated_records(
    labels: QuerySet | list | dict,
) -> list[str]:
    if not labels:
        return []
    registry = labels[0].__class__
    field = (
        REGISTRY_UNIQUE_FIELD.get(registry.__name__.lower(), "uid")
        if not hasattr(registry, "_ontology_id_field")
        else registry._ontology_id_field
    )
    # if the field value is None, use uid field
    label_uids = [getattr(label, field) for label in labels if label is not None]
    # save labels from ontology_ids
    if hasattr(registry, "_ontology_id_field") and label_uids:
        try:
            records = registry.from_values(label_uids, field=field, mute=True)
            save([r for r in records if r._state.adding])
        except Exception:  # noqa: S110
            pass
        field = "uid"
        label_uids = [label.uid for label in labels if label is not None]

    if issubclass(registry, CanCurate):
        validated = registry.validate(label_uids, field=field, mute=True)
        new_labels = [
            label for label, is_valid in zip(labels, validated) if not is_valid
        ]
        return new_labels
    return list(labels)


def save_validated_records(
    labels: QuerySet | list | dict,
) -> list[str] | dict[str, list[str]]:
    """Save validated labels from public based on ontology_id_fields."""
    if isinstance(labels, dict):
        return {
            registry: _save_validated_records(registry_labels)
            for registry, registry_labels in labels.items()
        }
    return _save_validated_records(labels)


class LabelManager:
    """Label manager.

    This allows to manage untyped labels :class:`~lamindb.ULabel` and arbitrary
    typed labels (e.g., :class:`~bionty.CellLine`) and associate labels
    with features.
    """

    def __init__(self, host: Artifact | Collection) -> None:
        self._host = host

    def __repr__(self) -> str:
        return self.describe(return_str=True)

    def describe(self, return_str=True) -> str:
        """Describe the labels."""
        tree = describe_labels(self._host)
        return format_rich_tree(
            tree, fallback="no linked labels", return_str=return_str
        )

    def add(
        self,
        records: SQLRecord | list[SQLRecord] | QuerySet,
        feature: Feature | None = None,
    ) -> None:
        """Add one or several labels and associate them with a feature.

        Args:
            records: Label records to add.
            feature: Feature under which to group the labels.
        """
        from .artifact import add_labels

        return add_labels(self._host, records=records, feature=feature)

    def get(
        self,
        feature: Feature,
        mute: bool = False,
        flat_names: bool = False,
    ) -> QuerySet | dict[str, QuerySet] | list:
        """Get labels given a feature.

        Args:
            feature: Feature under which labels are grouped.
            mute: Show no logging.
            flat_names: Flatten list to names rather than returning records.
        """
        from .artifact import get_labels

        return get_labels(self._host, feature=feature, mute=mute, flat_names=flat_names)

    def add_from(self, data: Artifact | Collection, transfer_logs: dict = None) -> None:
        """Add labels from an artifact or collection to another artifact or collection.

        Examples:
            >>> artifact1 = ln.Artifact(pd.DataFrame(index=[0, 1])).save()
            >>> artifact2 = ln.Artifact(pd.DataFrame(index=[2, 3])).save()
            >>> ulabels = ln.ULabel.from_values(["Label1", "Label2"], field="name")
            >>> ln.save(ulabels)
            >>> labels = ln.ULabel.filter(name__icontains = "label").all()
            >>> artifact1.ulabels.set(labels)
            >>> artifact2.labels.add_from(artifact1)
        """
        if transfer_logs is None:
            transfer_logs = {"mapped": [], "transferred": [], "run": None}
        from lamindb import settings

        using_key = settings._using_key
        for related_name, labels in _get_labels(data, instance=data._state.db).items():
            labels = labels.all()
            if not labels.exists():
                continue
            # look for features
            data_name_lower = data.__class__.__name__.lower()
            labels_by_features: dict = defaultdict(list)
            features = set()
            new_labels = save_validated_records(labels)
            if len(new_labels) > 0:
                transfer_fk_to_default_db_bulk(
                    new_labels, using_key, transfer_logs=transfer_logs
                )
            for label in labels:
                keys: list = []
                # if the link table doesn't follow this convention, we'll ignore it
                if not hasattr(label, f"links_{data_name_lower}"):
                    key = None
                    keys.append(key)
                else:
                    links = (
                        getattr(label, f"links_{data_name_lower}")
                        .filter(**{f"{data_name_lower}_id": data.id})
                        .all()
                    )
                    for link in links:
                        if link.feature is not None:
                            features.add(link.feature)
                            key = link.feature.uid
                        else:
                            key = None
                        keys.append(key)
                label_returned = transfer_to_default_db(
                    label,
                    using_key,
                    transfer_logs=transfer_logs,
                    transfer_fk=False,
                    save=True,
                )
                # TODO: refactor return value of transfer to default db
                if label_returned is not None:
                    label = label_returned
                for key in keys:
                    labels_by_features[key].append(label)
            # treat features
            new_features = save_validated_records(list(features))
            if len(new_features) > 0:
                transfer_fk_to_default_db_bulk(
                    new_features, using_key, transfer_logs=transfer_logs
                )
                for feature in new_features:
                    transfer_to_default_db(
                        feature,  # type: ignore
                        using_key,
                        transfer_logs=transfer_logs,
                        transfer_fk=False,
                    )
                save(new_features)  # type: ignore
            if hasattr(self._host, related_name):
                for feature_uid, feature_labels in labels_by_features.items():
                    if feature_uid is not None:
                        feature_id = Feature.get(feature_uid).id
                    else:
                        feature_id = None
                    getattr(self._host, related_name).add(
                        *feature_labels, through_defaults={"feature_id": feature_id}
                    )

    def make_external(self, label: SQLRecord) -> None:
        """Make a label external, aka dissociate label from internal features.

        Args:
            label: Label record to make external.
        """
        d = dict_related_model_to_related_name(self._host)
        registry = label.__class__
        related_name = d.get(registry.__get_name_with_module__())
        link_model = getattr(self._host, related_name).through
        link_records = link_model.filter(
            artifact_id=self._host.id, **{f"{registry.__name__.lower()}_id": label.id}
        )
        features = link_records.values_list("feature__name", flat=True).distinct()
        s = "s" if len(features) > 1 else ""
        link_records.update(feature_id=None, feature_ref_is_name=None)
        logger.warning(
            f'{registry.__name__} "{getattr(label, label._name_field)}" is no longer associated with the following feature{s}:\n'
            f"{list(features)}"
        )



# Generated by Django 5.1.7 on 2025-05-23 23:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0098_alter_feature_type_alter_project_type_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="writelog",
            name="seqno",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.RenameField(
            model_name="writelog",
            old_name="seqno",
            new_name="id",
        ),
    ]



"""Sub settings.

.. autosummary::
   :toctree: .

   CreationSettings
   AnnotationSettings

"""

from ._annotation_settings import AnnotationSettings
from ._creation_settings import CreationSettings



# Generated by Django 5.2 on 2025-05-25 11:59

import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields


def update_space_uids_and_create_branches(apps, schema_editor):
    Space = apps.get_model("lamindb", "Space")
    Space.objects.filter(uid="00000000").update(uid="A")
    Branch = apps.get_model("lamindb", "Branch")
    Branch.objects.get_or_create(
        id=-1,
        uid="T",
        name="Trash",
        description="The trash.",
    )
    Branch.objects.get_or_create(
        id=0,
        uid="A",
        name="Archive",
        description="The archive.",
    )
    Branch.objects.get_or_create(
        uid="M",
        name="Main",
        description="The main & default branch of the instance.",
    )


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0099_alter_writelog_seqno"),
    ]

    operations = [
        migrations.CreateModel(
            name="Branch",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(db_index=True, max_length=100)),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_default="M",
                        db_index=True,
                        default="M",
                        editable=False,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
        ),
        migrations.AlterField(
            model_name="space",
            name="uid",
            field=lamindb.base.fields.CharField(
                blank=True,
                db_default="A",
                db_index=True,
                default="A",
                editable=False,
                max_length=12,
                unique=True,
            ),
        ),
        migrations.RunPython(update_space_uids_and_create_branches),
    ]



import lamindb as ln
import pytest


def test_connect_reconnect():
    # testuser2 needs write access lamin-site-assets because of a fluke
    # in the legacy collaborator management, it seems
    assert ln.setup.settings.user.handle == "testuser2"
    ln.connect("lamindb-unit-tests-storage")  # this is not changing anything
    count1 = ln.Artifact.filter().count()
    # a public instance that does not have bionty configured
    ln.connect("laminlabs/lamin-site-assets")
    count2 = ln.Artifact.filter().count()
    assert count1 != count2
    with pytest.raises(ln.setup.errors.ModuleWasntConfigured):
        import bionty as bt
    ln.connect("lamindb-unit-tests-storage")
    import bionty as bt

    count3 = bt.Gene.filter().count()
    assert count2 != count3



import lamindb as ln
import pytest


def test_rename():
    import pandas as pd
    from lamindb.errors import SQLRecordNameChangeIntegrityError

    df = pd.DataFrame(
        {
            "feature_to_rename": [
                "label-to-rename",
                "label-to-rename",
                "label-not-to-rename",
            ],
            "feature_to_rename2": [
                "label-not-to-rename",
                "label-not-to-rename",
                "label-not-to-rename",
            ],
        }
    )

    curator = ln.Curator.from_dataframe(
        df,
        categoricals={
            "feature_to_rename": ln.ULabel.name,
            "feature_to_rename2": ln.ULabel.name,
        },
    )
    curator.add_new_from("feature_to_rename")
    curator.add_new_from("feature_to_rename2")
    artifact = curator.save_artifact(description="test-rename")
    assert artifact.ulabels.through.objects.filter(
        feature__name="feature_to_rename", ulabel__name="label-to-rename"
    ).exists()
    assert ln.Artifact.filter(feature_sets__features__name="feature_to_rename").exists()

    # rename label
    ulabel = ln.ULabel.get(name="label-to-rename")
    with pytest.raises(SQLRecordNameChangeIntegrityError):
        ulabel.name = "label-renamed"
        ulabel.save()

    artifact.labels.make_external(ulabel)
    assert not artifact.ulabels.through.objects.filter(
        feature__name="feature_to_rename", ulabel__name="label-to-rename"
    ).exists()
    ulabel.name = "label-renamed"
    ulabel.save()

    # rename feature
    feature = ln.Feature.get(name="feature_to_rename")
    with pytest.raises(SQLRecordNameChangeIntegrityError):
        feature.name = "feature_renamed"
        feature.save()

    artifact.features.make_external(feature)
    assert not ln.Artifact.filter(
        feature_sets__features__name="feature_to_rename"
    ).exists()
    assert ln.Artifact.filter(
        feature_sets__features__name="feature_to_rename2"
    ).exists()
    feature.name = "feature_renamed"
    feature.save()

    # rename the other feature, automatically deletes no-member schema
    feature2 = ln.Feature.get(name="feature_to_rename2")
    artifact.features.make_external(feature2)
    assert artifact.feature_sets.count() == 0

    # clean up
    artifact.delete(permanent=True)
    ln.Schema.filter().delete()
    ln.ULabel.filter().delete()
    ln.Feature.filter().delete()



# Generated by Django 5.2 on 2025-05-11 18:54

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0096_remove_artifact__param_values_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="ArtifactParamValue",
        ),
        migrations.DeleteModel(
            name="SchemaParam",
        ),
        migrations.DeleteModel(
            name="Param",
        ),
        migrations.DeleteModel(
            name="ParamValue",
        ),
        migrations.DeleteModel(
            name="RunParamValue",
        ),
    ]



# Generated by Django 5.2 on 2025-01-15 22:25

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
import django.db.models.functions.datetime
from django.db import migrations, models

import lamindb.base.fields
import lamindb.base.ids
import lamindb.base.users
import lamindb.models


class Migration(migrations.Migration):
    dependencies = [
        ("lamindb", "0077_lamindbv1_part6b"),
    ]

    operations = [
        migrations.CreateModel(
            name="TidyTable",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "updated_at",
                    lamindb.base.fields.DateTimeField(
                        blank=True,
                        db_default=django.db.models.functions.datetime.Now(),
                        db_index=True,
                        editable=False,
                    ),
                ),
                (
                    "_branch_code",
                    models.SmallIntegerField(db_default=1, db_index=True, default=1),
                ),
                (
                    "_aux",
                    lamindb.base.fields.JSONField(
                        blank=True, db_default=None, default=None, null=True
                    ),
                ),
                (
                    "uid",
                    lamindb.base.fields.CharField(
                        blank=True,
                        db_index=True,
                        default=lamindb.base.ids.base62_12,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    lamindb.base.fields.CharField(
                        blank=True, default=None, max_length=255
                    ),
                ),
                (
                    "is_type",
                    lamindb.base.fields.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "description",
                    lamindb.base.fields.TextField(blank=True, default=None),
                ),
                (
                    "created_by",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.base.users.current_user_id,
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.user",
                    ),
                ),
                (
                    "projects",
                    models.ManyToManyField(
                        related_name="_tidytables", to="lamindb.project"
                    ),
                ),
                (
                    "run",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        default=lamindb.models.current_run,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="lamindb.run",
                    ),
                ),
                (
                    "schema",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="lamindb.schema",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
                (
                    "type",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="records",
                        to="lamindb.tidytable",
                    ),
                ),
                (
                    "ulabels",
                    models.ManyToManyField(
                        related_name="_tidytables", to="lamindb.ulabel"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TidyTableData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "row",
                    lamindb.base.fields.IntegerField(
                        blank=True, help_text="Use -1 for result data"
                    ),
                ),
                ("value_int", models.BigIntegerField(blank=True, null=True)),
                ("value_float", models.FloatField(blank=True, null=True)),
                ("value_str", models.TextField(blank=True, null=True)),
                (
                    "value_upath",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("value_datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "value_json",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, null=True
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "param",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.param",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
                (
                    "tidytable",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="data",
                        to="lamindb.tidytable",
                    ),
                ),
                (
                    "value_artifact",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "value_collection",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "value_person",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.person",
                    ),
                ),
                (
                    "value_project",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.project",
                    ),
                ),
                (
                    "value_ulabel",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.ulabel",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RunData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "row",
                    lamindb.base.fields.IntegerField(
                        blank=True, help_text="Use -1 for result data"
                    ),
                ),
                ("value_int", models.BigIntegerField(blank=True, null=True)),
                ("value_float", models.FloatField(blank=True, null=True)),
                ("value_str", models.TextField(blank=True, null=True)),
                (
                    "value_upath",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("value_datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "value_json",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, null=True
                    ),
                ),
                (
                    "feature",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.feature",
                    ),
                ),
                (
                    "param",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.param",
                    ),
                ),
                (
                    "run",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="data",
                        to="lamindb.run",
                    ),
                ),
                (
                    "space",
                    lamindb.base.fields.ForeignKey(
                        blank=True,
                        db_default=1,
                        default=1,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="lamindb.space",
                    ),
                ),
                (
                    "value_artifact",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.artifact",
                    ),
                ),
                (
                    "value_collection",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.collection",
                    ),
                ),
                (
                    "value_person",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.person",
                    ),
                ),
                (
                    "value_project",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.project",
                    ),
                ),
                (
                    "value_ulabel",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lamindb.ulabel",
                    ),
                ),
            ],
            options={
                "indexes": [
                    models.Index(
                        fields=["run", "row"], name="lamindb_run_run_id_c02e71_idx"
                    ),
                    models.Index(
                        fields=["feature"], name="lamindb_run_feature_a37229_idx"
                    ),
                    models.Index(
                        fields=["param"], name="lamindb_run_param_i_cee058_idx"
                    ),
                ],
                "constraints": [
                    models.CheckConstraint(
                        condition=models.Q(
                            models.Q(
                                ("feature__isnull", False), ("param__isnull", True)
                            ),
                            models.Q(
                                ("feature__isnull", True), ("param__isnull", False)
                            ),
                            _connector="OR",
                        ),
                        name="run_data_feature_param_mutex",
                    ),
                    models.UniqueConstraint(
                        fields=("run", "row", "feature", "param"),
                        name="run_data_unique",
                    ),
                ],
            },
        ),
        migrations.AddIndex(
            model_name="tidytable",
            index=models.Index(fields=["uid"], name="lamindb_tid_uid_3a6e54_idx"),
        ),
        migrations.AddIndex(
            model_name="tidytable",
            index=models.Index(fields=["name"], name="lamindb_tid_name_50c5de_idx"),
        ),
        migrations.AddIndex(
            model_name="tidytabledata",
            index=models.Index(
                fields=["tidytable", "row"], name="lamindb_tid_tidytab_b35a4d_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="tidytabledata",
            index=models.Index(
                fields=["feature"], name="lamindb_tid_feature_5a0b1f_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="tidytabledata",
            index=models.Index(fields=["param"], name="lamindb_tid_param_i_16c884_idx"),
        ),
        migrations.AddConstraint(
            model_name="tidytabledata",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    models.Q(("feature__isnull", False), ("param__isnull", True)),
                    models.Q(("feature__isnull", True), ("param__isnull", False)),
                    _connector="OR",
                ),
                name="tidy_table_data_feature_param_mutex",
            ),
        ),
        migrations.AddConstraint(
            model_name="tidytabledata",
            constraint=models.UniqueConstraint(
                fields=("tidytable", "row", "feature", "param"),
                name="tidy_table_data_unique",
            ),
        ),
    ]



import shutil
from pathlib import Path
from subprocess import DEVNULL, run
from time import perf_counter

import lamindb as ln
import lamindb_setup as ln_setup
import pytest
from lamin_utils import logger
from laminci.db import setup_local_test_postgres


def create_test_instance(pgurl: str):
    ln.setup.init(
        storage="./default_storage_unit_storage",
        modules="bionty",
        name="lamindb-unit-tests-storage",
        db=pgurl,
    )
    ln.setup.register()  # temporarily
    ln.settings.creation.artifact_silence_missing_run_warning = True
    ln.settings.track_run_inputs = False
    ln.Storage("s3://lamindb-ci/test-data").save()
    ln.Storage("s3://lamindb-test/core").save()
    ln.Storage("s3://lamindb-test/storage").save()


def pytest_sessionstart():
    t_execute_start = perf_counter()

    ln_setup._TESTING = True
    try:
        pgurl = setup_local_test_postgres()
    except RuntimeError:
        run("docker stop pgtest && docker rm pgtest", shell=True, stdout=DEVNULL)  # noqa: S602
        pgurl = setup_local_test_postgres()
    try:
        create_test_instance(pgurl)
    except Exception as e:
        print("failed to create test instance:", e)
        print("deleting the instance")
        delete_test_instance()
        # below currently fails because cannot create two instances in the same session
        # create_test_instance(pgurl)
        print("now rerun")
        quit()
    total_time_elapsed = perf_counter() - t_execute_start
    print(f"time to setup the instance: {total_time_elapsed:.1f}s")
    assert ln.Storage.filter(root="s3://lamindb-ci/test-data").one_or_none() is not None


def delete_test_instance():
    logger.set_verbosity(1)
    if Path("./default_storage_unit_storage").exists():
        shutil.rmtree("./default_storage_unit_storage")
    # handle below better in the future
    for path in (
        "s3://lamindb-test/storage/.lamindb",
        "s3://lamindb-test/core/.lamindb",
        "s3://lamindb-ci/lamindb-unit-tests-cloud/.lamindb",
        "s3://lamindb-ci/test-settings-switch-storage/.lamindb",
    ):
        upath = ln.UPath(path)
        if upath.exists():
            upath.rmdir()
    ln.setup.delete("lamindb-unit-tests-storage", force=True)


def pytest_sessionfinish(session: pytest.Session):
    delete_test_instance()
    run("docker stop pgtest && docker rm pgtest", shell=True, stdout=DEVNULL)  # noqa: S602


@pytest.fixture
def ccaplog(caplog):
    """Add caplog handler to our custom logger at session start."""
    from lamin_utils._logger import logger

    # Add caplog's handler to our custom logger
    logger.addHandler(caplog.handler)

    yield caplog

    # Clean up at the end of the session
    logger.removeHandler(caplog.handler)



from __future__ import annotations

import importlib
import sys
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ... import Schema


def anndata_ensembl_gene_ids_and_valid_features_in_obs() -> Schema:
    """An `AnnData` schema validating Ensembl gene IDs and valid features in obs.

    .. literalinclude:: scripts/define_schema_anndata_ensembl_gene_ids_and_valid_features_in_obs.py
        :language: python
    """
    from ... import Schema

    docs_path = Path(__file__).parent.parent.parent.parent / "docs" / "scripts"
    if str(docs_path) not in sys.path:
        sys.path.append(str(docs_path))

    try:
        return Schema.get(name="anndata_ensembl_gene_ids_and_valid_features_in_obs")
    except Schema.DoesNotExist:
        import define_schema_anndata_ensembl_gene_ids_and_valid_features_in_obs  # noqa

        try:
            return Schema.get(name="anndata_ensembl_gene_ids_and_valid_features_in_obs")
        except Schema.DoesNotExist:
            importlib.reload(
                define_schema_anndata_ensembl_gene_ids_and_valid_features_in_obs
            )
            return Schema.get(name="anndata_ensembl_gene_ids_and_valid_features_in_obs")
